from __future__ import annotations

import weakref
import logging
from typing import TYPE_CHECKING, Iterable, Iterator, cast, Literal
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Signal, Qt, Property

from ._item_model import AbstractDataFrameModel
from ._header_view import QHorizontalHeaderView, QVerticalHeaderView
from ._table_base import QBaseTable, QMutableTable
from ._line_edit import QCellLiteralEdit, QCellLabelEdit

from tabulous._keymap import QtKeys
from tabulous._selection_model import RangesModel, SelectionModel, Index

if TYPE_CHECKING:
    from ._delegate import TableItemDelegate
    from tabulous._qt._mainwindow import _QtMainWidgetBase
    from tabulous._map_model import SlotRefMapping
    from tabulous._utils import TabulousConfig

# Flags
_SCROLL_PER_PIXEL = QtW.QAbstractItemView.ScrollMode.ScrollPerPixel

# Selection colors
CUR_COLOR = QtGui.QColor(128, 128, 128, 108)

logger = logging.getLogger("tabulous")


class MouseTrack:
    """Info about the mouse position and button state"""

    def __init__(self):
        self.last_rightclick_pos: QtCore.QPoint | None = None
        self.was_right_dragging: bool = False
        self.last_button: Literal["left", "right"] | None = None


class _EventFilter(QtCore.QObject):
    """An event filter for text completion by tab."""

    def eventFilter(self, o: _QTableViewEnhanced, e: QtCore.QEvent):
        _type = e.type()
        if _type == QtCore.QEvent.Type.KeyPress:
            e = cast(QtGui.QKeyEvent, e)
            if e.key() == Qt.Key.Key_Tab:
                if e.modifiers() == Qt.KeyboardModifier.NoModifier:
                    o._tab_clicked()
                    return True
        return False


class _QTableViewEnhanced(QtW.QTableView):
    selectionChangedSignal = Signal()
    rightClickedSignal = Signal(QtCore.QPoint)
    focusedSignal = Signal()
    resizedSignal = Signal()

    _focused_widget_ref: weakref.ReferenceType[QtW.QWidget] = None

    _table_map: SlotRefMapping

    def __init__(self, parent=None):
        super().__init__(parent)
        if isinstance(parent, QBaseTable):
            self._parent_table = parent
        else:
            self._parent_table = None

        self._zoom = 1.0
        self.setWordWrap(False)  # this disables eliding float text
        self.load_config()

        # use custom selection model
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.NoSelection)
        self._selection_model = SelectionModel(
            lambda: self.model().rowCount(),
            lambda: self.model().columnCount(),
        )
        self._selection_model.moving.connect(self._on_moving)
        self._selection_model.moved.connect(self._on_moved)
        self._highlight_model = RangesModel()

        # parameters for mouse tracking
        self._mouse_track = MouseTrack()

        # header settings
        vheader = QVerticalHeaderView()
        hheader = QHorizontalHeaderView()
        self.setVerticalHeader(vheader)
        self.setHorizontalHeader(hheader)

        vheader.resize(36, vheader.height())
        vheader.setMinimumSectionSize(5)
        hheader.setMinimumSectionSize(5)

        vheader.setDefaultSectionSize(self._h_default)
        hheader.setDefaultSectionSize(self._w_default)

        self.setVerticalScrollMode(_SCROLL_PER_PIXEL)
        self.setHorizontalScrollMode(_SCROLL_PER_PIXEL)

        # item delegate
        from ._delegate import TableItemDelegate

        delegate = TableItemDelegate(parent=self)
        self.setItemDelegate(delegate)
        self._update_all()

        # event filter
        self._event_filter = _EventFilter(self)
        self.installEventFilter(self._event_filter)

        # the source ranges of in-cell slot are drawed or not
        self._current_drawing_slot_ranges = None

        # initialize with dummy mapping
        from tabulous._map_model import DummySlotRefMapping

        self._table_map = DummySlotRefMapping()

        # initialize colors
        self._selection_color = QtGui.QColor(120, 120, 170, 255)
        self._highlight_color = QtGui.QColor(255, 0, 0, 86)

    # fmt: off
    if TYPE_CHECKING:
        def model(self) -> AbstractDataFrameModel: ...
        def itemDelegate(self) -> TableItemDelegate: ...
        def verticalHeader(self) -> QVerticalHeaderView: ...
        def horizontalHeader(self) -> QHorizontalHeaderView: ...
    # fmt: on

    def load_config(self, cfg: TabulousConfig | None = None) -> None:
        from tabulous._utils import get_config

        if cfg is None:
            cfg = get_config()

        table = cfg.table

        # settings
        self._font_size = table.font_size
        self._h_default = table.row_size
        self._w_default = table.column_size
        self._font = table.font
        qfont = QtGui.QFont(self._font, int(self._zoom * self._font_size))
        self.setFont(qfont)
        self.horizontalHeader().setFont(qfont)
        self.verticalHeader().setFont(qfont)
        self._update_all()

    @property
    def _focused_widget(self) -> QtW.QWidget | None:
        """QWidget that force focusing after focus is moved to the table."""
        if self.__class__._focused_widget_ref is None:
            return None
        return self.__class__._focused_widget_ref()

    @_focused_widget.setter
    def _focused_widget(self, widget: QtW.QWidget | None) -> None:
        current = self._focused_widget

        if widget is None:
            self.__class__._focused_widget_ref = None
        else:
            self.__class__._focused_widget_ref = weakref.ref(widget)

        if current is not None:
            current.close()

    @_focused_widget.deleter
    def _focused_widget(self) -> None:
        self._focused_widget = None

    def _update_all(self, rect: QtCore.QRect | None = None) -> None:
        """repaint the table and the headers."""
        if rect is None:
            self.viewport().update()
        else:
            rect.adjust(-2, -2, 2, 2)
            self.viewport().update(rect)
        self.horizontalHeader().viewport().update()
        self.verticalHeader().viewport().update()
        return None

    def _update_rect(self, rect: QtCore.QRect) -> None:
        rect.adjust(-2, -2, 2, 2)
        return self.viewport().update(rect)

    def _range_rect(self, rng: tuple[slice, slice]) -> QtCore.QRect:
        rsel, csel = rng
        model = self.model()
        rect = self.visualRect(model.index(rsel.start, csel.start))
        rect |= self.visualRect(model.index(rsel.stop - 1, csel.stop - 1))
        return rect

    def _tab_clicked(self) -> None:
        r, c = self._selection_model.current_index
        if c == self.model().columnCount() - 1:
            if r < self.model().rowCount() - 1:
                self._selection_model.move_to(r + 1, 0)
            else:
                return None
        else:
            self._selection_model.move(0, 1)
        return None

    def _on_moving(self, src: Index, dst: Index) -> None:
        _need_update_all = self._current_drawing_slot_ranges is not None

        _nr, _nc = self.parentTable().dataShape()
        _r0, _c0 = dst
        new_status_tip = ""
        if _r0 < _nr and _c0 < _nc:
            _r0 = self.parentTable()._proxy.get_source_index(_r0)
            if slot := self._table_map.get_by_dest((_r0, _c0), None):
                self._current_drawing_slot_ranges = slot.range
                new_status_tip = f"<b><code>{slot.as_literal(dest=True)}</code></b>"
                _need_update_all = True
            else:
                self._current_drawing_slot_ranges = None

        if qviewer := self.parentViewer():
            qviewer._table_viewer.status = new_status_tip

        if _need_update_all:
            self._update_all()
            return None

        if not self._selection_model.is_jumping():
            # clear all the multi-selections
            for sel in self._selection_model:
                self._update_rect(self._range_rect(sel))

        else:
            if len(self._selection_model) > 1:
                self._update_rect(self._range_rect(self._selection_model[-2]))

        if self._selection_model.is_moving_to_edge():
            if len(self._selection_model) > 0:
                self._update_rect(self._range_rect(self._selection_model[-1]))

        return None

    def _on_moved(self, src: Index, dst: Index) -> None:
        """Update the view."""
        model = self.model()
        index_src = model.index(*src.as_uint())
        index_dst = model.index(*dst.as_uint())
        if dst >= (0, 0) and self.hasFocus():
            self.scrollTo(index_dst)

        # rect is the region that needs to be updated
        rect: QtCore.QRect = self.visualRect(index_dst)
        if not self._selection_model.is_jumping():
            rect |= self.visualRect(index_src)
        if sel := self._selection_model.current_range:
            rect |= self._range_rect(sel)
        if start := self._selection_model.start:
            rect |= self.visualRect(model.index(*start))

        if src.row < 0 or dst.row < 0:
            rect.setBottom(999999)
        if src.column < 0 or dst.column < 0:
            rect.setRight(999999)

        self.selectionChangedSignal.emit()

        self._update_all(rect)
        return None

    def copy(self, link: bool = True) -> _QTableViewEnhanced:
        """Make a copy of the table."""
        new = _QTableViewEnhanced(self.parentTable())
        if link:
            new.setModel(self.model())
            new._selection_model = self._selection_model
            new._selection_model.moving.connect(new._on_moving)
            new._selection_model.moved.connect(new._on_moved)
            new._table_map = self._table_map
        new.setZoom(self.zoom())
        new._selection_model.current_index = self._selection_model.current_index
        return new

    def selectAll(self) -> None:
        """Override selectAll slot to update custom selections."""
        nr, nc = self.model().df.shape
        if nr * nc > 0:
            self.set_selections([(slice(0, nr), slice(0, nc))])
        return None

    def clear_selections(self) -> None:
        """Clear current selections."""
        self._selection_model.clear()
        self._update_all()
        return None

    def set_selections(self, selections: list[tuple[slice, slice]]) -> None:
        """Set current selections."""
        self._selection_model.set_ranges(selections)
        self.selectionChangedSignal.emit()
        self._update_all()
        return None

    def clear_highlights(self) -> None:
        """Clear current highlights."""
        self._highlight_model.clear()
        self._update_all()
        return None

    def set_highlights(self, highlights: list[tuple[slice, slice]]) -> None:
        """Set current highlights."""
        self._highlight_model.set_ranges(highlights)
        self._update_all()
        return None

    def _edit_current(self) -> QCellLiteralEdit | None:
        """Enter edit mode for current cell."""
        index = self.model().index(*self._selection_model.current_index)
        self.edit(index)
        if editor := self._focused_widget:
            if isinstance(editor, QCellLiteralEdit):
                editor = cast(QCellLiteralEdit, editor)
                editor._on_text_changed(editor.text())
                editor._self_focused = True
            editor.setFocus()
        return editor

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        """Register clicked position"""
        # initialize just in case

        _selection_model = self._selection_model
        _selection_model.set_ctrl(e.modifiers() & Qt.KeyboardModifier.ControlModifier)
        self._mouse_track.last_rightclick_pos = e.pos()
        if e.button() == Qt.MouseButton.LeftButton:
            index = self.indexAt(e.pos())
            if index.isValid():
                r, c = index.row(), index.column()
                self._selection_model.jump_to(r, c)
            else:
                # if outside the table is clicked, close the editor
                if isinstance(self._focused_widget, QCellLiteralEdit):
                    line = cast(QCellLiteralEdit, self._focused_widget)
                    line.eval_and_close()
            self._mouse_track.last_button = "left"
        elif e.button() == Qt.MouseButton.RightButton:
            self._mouse_track.was_right_dragging = False
            self._mouse_track.last_button = "right"
            return
        _selection_model.set_shift(True)
        return super().mousePressEvent(e)

    def mouseMoveEvent(self, e: QtGui.QMouseEvent) -> None:
        """Scroll table plane when mouse is moved with right click."""
        if self._mouse_track.last_button == "right":
            pos = e.pos()
            dy = pos.y() - self._mouse_track.last_rightclick_pos.y()
            dx = pos.x() - self._mouse_track.last_rightclick_pos.x()
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() - dy)
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - dx)
            self._mouse_track.last_rightclick_pos = pos
            self._mouse_track.was_right_dragging = True
        else:
            index = self.indexAt(e.pos())
            if index.isValid():
                r, c = index.row(), index.column()
                if self._selection_model.current_index != (r, c):
                    self._selection_model.move_to(r, c)

        return None

    def mouseReleaseEvent(self, e: QtGui.QMouseEvent) -> None:
        """Delete last position."""
        if (
            e.button() == Qt.MouseButton.RightButton
            and not self._mouse_track.was_right_dragging
        ):
            self.rightClickedSignal.emit(e.pos())
        self._mouse_track.last_rightclick_pos = None
        self._mouse_track.last_button = None
        self._selection_model.set_shift(
            e.modifiers() & Qt.KeyboardModifier.ShiftModifier
        )
        self._mouse_track.was_right_dragging = False
        if wdt := self._focused_widget:
            wdt.setFocus()
        return super().mouseReleaseEvent(e)

    def mouseDoubleClickEvent(self, e: QtGui.QMouseEvent) -> None:
        index = self.indexAt(e.pos())
        if not index.isValid():
            return None
        from ._table_base import QMutableTable

        table = self.parentTable()
        if isinstance(table, QMutableTable):
            if table.isEditable():
                self.edit(index)
                self._selection_model.set_shift(False)
            else:
                if stack := table.tableStack():
                    stack.notifyEditability()
        return None

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        """Evoke parent keyPressEvent."""
        keys = QtKeys(e)

        sel_mod = self._selection_model

        if keys.has_shift():
            sel_mod.set_shift(True)
        else:
            sel_mod.set_shift(False)
            if keys.has_key():
                sel_mod.reset()

        parent = self.parentTable()

        if keys.is_typing() and parent.isEditable():
            # First check if either header is selected. If not, then edit
            # the current table cell.
            parent = cast(QMutableTable, parent)
            sel_mod.set_shift(False)
            sel_mod.reset()

            if sel_mod.current_index.row < 0:
                focused_widget = parent.editHorizontalHeader(
                    sel_mod.current_index.column
                )

            elif sel_mod.current_index.column < 0:
                focused_widget = parent.editVerticalHeader(sel_mod.current_index.row)

            else:
                focused_widget = self._edit_current()

            if isinstance(focused_widget, QtW.QLineEdit):
                focused_widget = cast(QtW.QLineEdit, focused_widget)
                focused_widget.setText(keys.key_string(check_shift=True))
                focused_widget.deselect()

            return None

        elif keys == "F2":
            if not parent.isEditable():
                if stack := parent.tableStack():
                    stack.notifyEditability()
                return None
            parent = cast(QMutableTable, parent)

            if sel_mod.current_index.row < 0:
                parent.editHorizontalHeader(sel_mod.current_index.column)
            elif sel_mod.current_index.column < 0:
                parent.editVerticalHeader(sel_mod.current_index.row)
            else:
                self._edit_current()

            return None

        elif keys == "F3":
            editor = QCellLabelEdit.from_table(self)
            editor.show()
            return None

        if keys.has_ctrl():
            sel_mod.set_ctrl(True)
        elif keys.has_key():
            sel_mod.set_ctrl(False)

        if isinstance(parent, QBaseTable):
            return parent.keyPressEvent(e)

    def keyReleaseEvent(self, a0: QtGui.QKeyEvent) -> None:
        keys = QtKeys(a0)
        self._selection_model.set_ctrl(keys.has_ctrl())
        self._selection_model.set_shift(
            keys.has_shift() or self._mouse_track.last_rightclick_pos is not None
        )
        return super().keyReleaseEvent(a0)

    def inputMethodEvent(self, event: QtGui.QInputMethodEvent) -> None:
        """Catch Japanese/Chinese edit event."""
        # NOTE: super().inputMethodEvent(event) is buggy!!
        if event.preeditString():
            self._edit_current()
            # FIXME: Cannot send the pre-edit text to the editor.
            # editor.inputMethodQuery(Qt.InputMethodQuery.ImSurroundingText)

    def zoom(self) -> float:
        """Get current zoom factor."""
        return self._zoom

    def setZoom(self, value: float) -> None:
        """Set zoom factor."""
        if not 0.25 <= value <= 2.0:
            raise ValueError("Zoom factor must between 0.25 and 2.0.")
        # To keep table at the same position.
        zoom_ratio = 1 / self.zoom() * value
        pos = self.verticalScrollBar().sliderPosition()
        self.verticalScrollBar().setSliderPosition(int(pos * zoom_ratio))
        pos = self.horizontalScrollBar().sliderPosition()
        self.horizontalScrollBar().setSliderPosition(int(pos * zoom_ratio))

        # Zoom section size of headers
        self.setSectionSize(int(self._w_default * value), int(self._h_default * value))

        # Update stuff
        self._zoom = value
        font = self.font()
        font.setPointSize(int(self._font_size * value))
        self.setFont(font)
        self.verticalHeader().setFont(font)
        self.horizontalHeader().setFont(font)
        self._update_all()
        return

    def wheelEvent(self, e: QtGui.QWheelEvent) -> None:
        """Zoom in/out table."""
        if e.modifiers() & Qt.KeyboardModifier.ControlModifier:
            dt = e.angleDelta().y() / 120
            zoom = self.zoom() + 0.15 * dt
            self.setZoom(min(max(zoom, 0.25), 2.0))
            return None

        return super().wheelEvent(e)

    def sectionSize(self) -> tuple[int, int]:
        """Return current section size."""
        return (
            self.horizontalHeader().defaultSectionSize(),
            self.verticalHeader().defaultSectionSize(),
        )

    def setSectionSize(self, horizontal: int, vertical: int) -> None:
        """Update section size of headers."""
        self.verticalHeader().setDefaultSectionSize(vertical)
        self.horizontalHeader().setDefaultSectionSize(horizontal)
        return

    def focusInEvent(self, e: QtGui.QFocusEvent) -> None:
        self.focusedSignal.emit()
        # initialize selection model state
        self._selection_model.set_ctrl(False)
        self._selection_model.set_shift(False)
        return super().focusInEvent(e)

    def resizeEvent(self, e: QtGui.QResizeEvent) -> None:
        self.resizedSignal.emit()
        return super().resizeEvent(e)

    def _get_selection_color(self):
        return self._selection_color

    def _set_selection_color(self, value: QtGui.QColor):
        self._selection_color = value
        self.update()

    selectionColor = Property(QtGui.QColor, _get_selection_color, _set_selection_color)

    def _get_highlight_color(self):
        return self._highlight_color

    def _set_highlight_color(self, value: QtGui.QColor):
        self._highlight_color = value
        self.update()

    highlightColor = Property(QtGui.QColor, _get_highlight_color, _set_highlight_color)

    def _get_current_index_color(self):
        return CUR_COLOR

    def paintEvent(self, event: QtGui.QPaintEvent):
        """Paint table and the selection."""
        super().paintEvent(event)
        focused = int(self.hasFocus())
        nsel = len(self._selection_model)
        painter = QtGui.QPainter(self.viewport())

        # draw highlights
        h_color = self._get_highlight_color()
        for i, rect in enumerate(
            self._rect_from_ranges(self._highlight_model._ranges, map=True)
        ):
            painter.fillRect(rect, h_color)

        # draw selections
        s_color = self._get_selection_color()
        for i, rect in enumerate(self._rect_from_ranges(self._selection_model._ranges)):
            last_one = nsel == i + 1
            pen = QtGui.QPen(s_color, 2 + int(last_one) * focused)
            painter.setPen(pen)
            painter.drawRect(rect)

        # current index
        idx = self._selection_model.current_index
        if idx >= (0, 0):
            rect_cursor = self.visualRect(self.model().index(*idx))
            rect_cursor.adjust(1, 1, -1, -1)
            pen = QtGui.QPen(CUR_COLOR, 3)
            painter.setPen(pen)
            painter.drawRect(rect_cursor)

        # in-cell slot source ranges of the current index
        color_cycle = _color_cycle()
        if rng := self._current_drawing_slot_ranges:
            for rect in self._rect_from_ranges(rng.iter_ranges(), map=True):
                rect.adjust(1, 1, -1, -1)
                pen = QtGui.QPen(next(color_cycle), 3)
                painter.setPen(pen)
                painter.drawRect(rect)

        return None

    def parentTable(self) -> QBaseTable | None:
        """The parent QBaseTable widget."""
        parent = self._parent_table
        if not isinstance(parent, QBaseTable):
            parent = None
        return parent

    def parentViewer(self) -> _QtMainWidgetBase | None:
        """The parent table viewer widget."""
        parent = self.parentTable()
        return parent.parentViewer()

    def _rect_from_ranges(
        self,
        ranges: Iterable[tuple[slice, slice]],
        map: bool = False,
    ) -> Iterator[QtCore.QRect]:
        """Convert range models into rectangles."""
        model = self.model()
        prx = self.parentTable()._proxy
        for rr, cc in ranges:
            if map:
                # can only draw single-row selections during unordered state
                try:
                    rr = prx.map_slice(rr)
                except Exception:
                    continue
            top_left = model.index(rr.start, cc.start)
            bottom_right = model.index(rr.stop - 1, cc.stop - 1)
            rect = self.visualRect(top_left) | self.visualRect(bottom_right)
            yield rect

    def _create_eval_editor(
        self, text: str | None = None, moveto: tuple[int, int] | None = None
    ) -> QCellLiteralEdit:
        if moveto is not None:
            self._selection_model.move_to(*moveto)
        index = self.model().index(*self._selection_model.current_index)
        if text is None:
            text = self.model().data(index, Qt.ItemDataRole.EditRole)
        if not isinstance(text, str):
            text = ""
        line = QCellLiteralEdit.from_table(self, text)
        line.show()
        line.setFocus()
        return line


def _color_cycle() -> Iterator[QtGui.QColor]:
    """Generate easily distinguishable colors."""
    # This is the default color cycle of matplotlib
    colors = [
        QtGui.QColor(31, 119, 180),
        QtGui.QColor(255, 127, 14),
        QtGui.QColor(44, 160, 44),
        QtGui.QColor(214, 39, 40),
        QtGui.QColor(148, 103, 189),
        QtGui.QColor(140, 86, 75),
        QtGui.QColor(227, 119, 194),
        QtGui.QColor(127, 127, 127),
        QtGui.QColor(188, 189, 34),
        QtGui.QColor(23, 190, 207),
    ]
    ncolor = len(colors)
    i = 0
    while True:
        yield colors[i]
        i = i + 1 if i < ncolor - 1 else 0



from __future__ import annotations
import re
from typing import TYPE_CHECKING, Literal, cast, overload
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Signal, Qt

if TYPE_CHECKING:
    from .._mainwindow import _QtMainWidgetBase
    from tabulous.widgets import TableBase, Table, SpreadSheet


def create_temporal_line_edit(
    rect: QtCore.QRect,
    parent: QtW.QWidget,
    text: str,
) -> QtW.QLineEdit:
    line = QtW.QLineEdit(parent=parent)
    geometry = line.geometry()
    geometry.setWidth(rect.width())
    geometry.setHeight(rect.height())
    geometry.moveCenter(rect.center())
    geometry.adjust(4, 4, -2, -2)
    line.setGeometry(geometry)
    line.setText(text)
    line.setHidden(False)
    line.setFocus()
    line.selectAll()
    return line


def find_parent_table_viewer(qwidget: _QtMainWidgetBase) -> _QtMainWidgetBase:
    x = qwidget
    while (parent := x.parent()) is not None:
        x = parent
        if hasattr(x, "_table_viewer"):
            return x
    raise RuntimeError


_OPERATORS = re.compile(r"\+|-|\*|/|\s|\(|\)|%|<|>|=|\"|\'")


class _EventFilter(QtCore.QObject):
    """An event filter for text completion by tab."""

    def eventFilter(self, o: QtCore.QObject, e: QtCore.QEvent):
        if e.type() == QtCore.QEvent.Type.KeyPress:
            e = cast(QtGui.QKeyEvent, e)
            if e.key() == Qt.Key.Key_Tab:
                l = cast(QCompletableLineEdit, self.parent())
                l.setSelection(len(l.text()), len(l.text()))
                return True
        return False


class QCompletableLineEdit(QtW.QLineEdit):
    """A QLineEdit implemented with completion."""

    enterClicked = Signal()
    escClicked = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        from tabulous._utils import get_config

        table_config = get_config().table
        font = QtGui.QFont(table_config.font, table_config.font_size)
        self.setFont(font)
        self._qtable_viewer = find_parent_table_viewer(self)
        self.textChanged.connect(self.setCompletion)
        self._history: list[str] = []
        self._history_pos: int = -1
        self._event_filter = _EventFilter(self)
        self.installEventFilter(self._event_filter)
        self._last_key = None

    def currentQTable(self):
        tablestack = self._qtable_viewer._tablestack
        idx = tablestack.currentIndex()
        return tablestack.tableAtIndex(idx)

    @overload
    def currentPyTable(self, assert_mutable: Literal[False] = False) -> TableBase:
        ...

    @overload
    def currentPyTable(
        self, assert_mutable: Literal[True] = False
    ) -> Table | SpreadSheet:
        ...

    def currentPyTable(self, assert_mutable: bool = False):
        viewer = self._qtable_viewer._table_viewer
        table = viewer.current_table
        if assert_mutable and table.table_type not in ("Table", "SpreadSheet"):
            raise TypeError(f"Table {table.name} is immutable.")
        return viewer.current_table

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        key = event.key()
        self._last_key = key
        if key == Qt.Key.Key_Return:
            self.enterClicked.emit()
        elif key == Qt.Key.Key_Escape:
            self.escClicked.emit()
        elif key == Qt.Key.Key_Tab:
            self.setSelection(len(self.text()), len(self.text()))
        elif key == Qt.Key.Key_Up and self._history:
            self._history_pos -= 1
            if self._history_pos < 0:
                self._history_pos = 0
            self.setText(self._history[self._history_pos])

        elif key == Qt.Key.Key_Down:
            if self._history_pos == len(self._history):
                return
            self._history_pos += 1
            if self._history_pos == len(self._history):
                self.setText("")
            else:
                self.setText(self._history[self._history_pos])
        else:
            super().keyPressEvent(event)

    def toHistory(self):
        self._history.append(self.text())
        if len(self._history) > 300:
            self._history = self._history[-300:]
        self.setText("")
        self._history_pos = len(self._history)

    def setCompletion(self, text: str):
        """Set auto completion for the text in the line edit."""
        if self._last_key in (Qt.Key.Key_Backspace, Qt.Key.Key_Delete):
            # don't autocomplete if the user deleted a character
            return

        elif self.cursorPosition() < len(self.text()):
            return

        new_text = _complete(text, self.currentPyTable().data.columns)
        if new_text is not None:
            self.blockSignals(True)
            self.setText(new_text)
            self.blockSignals(False)
            self.setCursorPosition(len(text))
            self.setSelection(len(text), len(new_text))


def _complete(text: str, words: list) -> str | None:
    last_word = _OPERATORS.split(text)[-1].strip()
    if not last_word.isidentifier():
        return
    matched: list[str] = []
    for name in words:
        name = str(name)
        if name.startswith(last_word):
            matched.append(name)

    if not matched:
        return None

    matched.sort()
    _to_complete = matched[0][len(last_word) :]
    return text + _to_complete



from __future__ import annotations

from pathlib import Path
import weakref
from typing import TYPE_CHECKING, cast
from contextlib import suppress

from qtpy.QtCore import Signal
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtconsole.rich_jupyter_widget import RichJupyterWidget

from tabulous._keymap import QtKeys, QtKeyMap

if TYPE_CHECKING:
    from tabulous._qt._dockwidget import QtDockWidget
    from tabulous.widgets._mainwindow import TableViewerBase

    class RichJupyterWidget(RichJupyterWidget, QtW.QWidget):
        """To fix typing problem"""


# Modified from napari_console https://github.com/napari/napari-console
class QtConsole(RichJupyterWidget):
    codeExecuted = Signal(str)
    _keymap = QtKeyMap()

    def __init__(self, *args, **kwargs):
        self._old_point = None
        super().__init__(*args, **kwargs)
        self.setMinimumSize(100, 0)
        self.resize(100, 40)
        self._dock_parent = None
        self.codeExecuted.connect(self.setFocus)

    def connect_parent(self, widget: TableViewerBase):
        from IPython import get_ipython
        from IPython.terminal.interactiveshell import TerminalInteractiveShell
        from ipykernel.connect import get_connection_file
        from ipykernel.inprocess.ipkernel import InProcessInteractiveShell
        from ipykernel.zmqshell import ZMQInteractiveShell
        from qtconsole.client import QtKernelClient
        from qtconsole.inprocess import QtInProcessKernelManager

        shell = get_ipython()

        if shell is None:
            # If there is no currently running instance create an in-process
            # kernel.
            kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel(show_banner=False)
            kernel_manager.kernel.gui = "qt"

            kernel_client = kernel_manager.client()
            kernel_client.start_channels()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell: InProcessInteractiveShell = kernel_manager.kernel.shell
            self.push = self.shell.push

        elif type(shell) == InProcessInteractiveShell:
            # If there is an existing running InProcessInteractiveShell
            # it is likely because multiple viewers have been launched from
            # the same process. In that case create a new kernel.
            # Connect existing kernel
            kernel_manager = QtInProcessKernelManager(kernel=shell.kernel)
            kernel_client = kernel_manager.client()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push

        elif isinstance(shell, TerminalInteractiveShell):
            # if launching from an ipython terminal then adding a console is
            # not supported. Instead users should use the ipython terminal for
            # the same functionality.
            self.kernel_client = None
            self.kernel_manager = None
            self.shell = None
            self.push = lambda var: None

        elif isinstance(shell, ZMQInteractiveShell):
            # if launching from jupyter notebook, connect to the existing
            # kernel
            kernel_client = QtKernelClient(connection_file=get_connection_file())
            kernel_client.load_connection_file()
            kernel_client.start_channels()

            self.kernel_manager = None
            self.kernel_client = kernel_client
            self.shell = shell
            self.push = self.shell.push
        else:
            raise ValueError(f"ipython shell not recognized, got {type(shell)}")

        if self.shell is not None:

            from IPython.paths import get_ipython_dir
            from tabulous._utils import get_config

            config = get_config()
            _ns = config.console_namespace

            # run IPython startup files
            profile_dir = Path(get_ipython_dir()) / "profile_default" / "startup"
            if profile_dir.exists() and _ns.load_startup_file:
                import runpy

                _globals = {}
                for startup in profile_dir.glob("*.py"):
                    with suppress(Exception):
                        _globals.update(runpy.run_path(startup))

                self.shell.push(_globals)

            # update namespaces
            import tabulous as tbl
            import numpy as np
            import pandas as pd

            ns = {
                _ns.viewer: widget,
                _ns.numpy: np,
                _ns.pandas: pd,
                _ns.tabulous: tbl,
            }
            self.shell.push(ns)

    def setFocus(self):
        """Set focus to the text edit."""
        self._control.setFocus()
        cursor = self._control.textCursor()
        cursor.clearSelection()
        self._control.setTextCursor(cursor)
        return None

    def isActive(self) -> bool:
        return self.isVisible() and self.shell is not None

    def buffer(self) -> str:
        """Get current code block"""
        return self.input_buffer

    def setBuffer(self, code: str) -> None:
        """Set code string to Jupyter QtConsole buffer"""
        self.input_buffer = ""
        if not isinstance(code, str):
            raise ValueError(f"Cannot set {type(code)}.")
        cursor = self._control.textCursor()
        lines = code.split("\n")
        for line in lines[:-1]:
            cursor.insertText(line + "\n")
            self._insert_continuation_prompt(cursor)  # insert "...:"
        cursor.insertText(lines[-1])
        return None

    def insertText(self, text: str) -> None:
        cursor = self._control.textCursor()
        cursor.insertText(text)
        return None

    def setTempText(self, text: str | None = None) -> None:
        if text is None:
            text = "viewer.data.loc[...]"
        cursor = self._control.textCursor()
        cursor.removeSelectedText()
        pos = cursor.position()
        cursor.insertText(text)
        cursor.setPosition(pos)
        cursor.setPosition(pos + len(text), QtGui.QTextCursor.MoveMode.KeepAnchor)
        self._control.setTextCursor(cursor)
        return None

    def selectedText(self) -> str:
        """Return the selected text"""
        cursor = self._control.textCursor()
        return cursor.selection().toPlainText()

    def execute(
        self,
        source: str | None = None,
        hidden: bool = False,
        interactive: bool = False,
    ):
        """Execute current code block."""
        if source is None:
            source = self.input_buffer
        super().execute(source=source, hidden=hidden, interactive=interactive)
        self.codeExecuted.emit(source)
        return None

    # NOTE: qtconsole overwrites "parent" method so we have to use another method to
    # manage parent.
    def dockParent(self) -> QtDockWidget:
        """Return the dock widget parent."""
        if self._dock_parent is None:
            return None
        return self._dock_parent()

    def setDockParent(self, widget: QtDockWidget):
        """Set the dock widget parent."""
        if not isinstance(widget, QtW.QDockWidget):
            raise TypeError("Parent must be a QDockWidget")
        self._dock_parent = weakref.ref(widget)

    def eventFilter(self, obj, event: QtCore.QEvent):
        """Reimplemented to ensure a console-like behavior in the underlying
        text widgets.
        """
        etype = event.type()
        if etype == QtCore.QEvent.Type.KeyPress:
            parent = self.dockParent()
            if parent is None:
                return super().eventFilter(obj, event)

            event = cast(QtGui.QKeyEvent, event)
            keys = QtKeys(event)
            if self._keymap.press_key(keys):
                return True

        return super().eventFilter(obj, event)

    def update_console(self, ns: dict) -> None:
        """Update the console namespace."""
        self.shell.push(dict(ns))
        return None

    @_keymap.bind("Ctrl+Shift+Down", floating=False)
    @_keymap.bind("Ctrl+Shift+Up", floating=True)
    def setDockFloating(self, floating: bool):
        """Make the parent dock widget floating."""
        parent = self.dockParent()
        if parent is None:
            return
        if floating:
            parent.setFloating(True)
            QtCore.QTimer.singleShot(0, parent.activateWindow)
        else:
            parent.setFloating(False)
            self._control.ensureCursorVisible()
            self._control.setTextCursor(self._control.textCursor())

        if floating:
            if self._old_point is None:
                _screen_rect = QtGui.QGuiApplication.primaryScreen().geometry()
                _screen_center = _screen_rect.center()
                parent.resize(500, 600)
                pos = _screen_center - self.rect().center()
            else:
                pos = self._old_point
            parent.move(pos)
            self._old_point = pos
        parent.setFocus()
        self.setFocus()
        return None

    @_keymap.bind("Ctrl+Shift+C", keys="Ctrl+Shift+C")
    @_keymap.bind("Ctrl+Shift+F", keys="Ctrl+Shift+F")
    def _press_parent_keycombo(self, keys: str):
        parent = self.dockParent()
        if parent is None:
            return
        parent.parentWidget()._keymap.press_key(keys)

    @_keymap.bind("Ctrl+.")
    def _ignore(self):
        """Ignore restarting the kernel."""
        parent = self.parentWidget()
        while parent is not None:
            parent = parent.parentWidget()
            if hasattr(parent, "_table_viewer"):
                parent._keymap.press_key("Ctrl+.")
                break

    def update_theme(self, theme: str):
        from tabulous.style import Style
        from tabulous.color import normalize_color

        style = Style.from_global(theme)
        stylesheet = style.format_file()
        self.style_sheet = stylesheet

        # Set syntax styling and highlighting using theme
        light_theme = sum(normalize_color(style.background)[:3]) > 382.5
        if light_theme:
            self.syntax_style = "default"
        else:
            self.syntax_style = "vim"
        bracket_color = QtGui.QColor(*normalize_color(style.highlight0))
        self._bracket_matcher.format.setBackground(bracket_color)



from __future__ import annotations
from functools import wraps
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui

try:
    import matplotlib as mpl
except ImportError as e:
    raise ImportError(
        "Module 'matplotlib' is not installed. Please install it to use plot canvas."
    )
import matplotlib.pyplot as plt
from ._mpl_canvas import InteractiveFigureCanvas

if TYPE_CHECKING:
    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure
    from matplotlib.axes import Axes
    from matplotlib.artist import Artist

    import seaborn as sns
    from seaborn.axisgrid import Grid


class QtMplPlotCanvas(QtW.QWidget):
    """A matplotlib figure canvas."""

    _current_widget: QtMplPlotCanvas | None = None

    def __init__(
        self,
        nrows=1,
        ncols=1,
        style=None,
    ):
        backend = mpl.get_backend()
        try:
            mpl.use("Agg")
            if style is None:
                fig, _ = plt.subplots(nrows, ncols, linewidth=2)
            else:
                with plt.style.context(style):
                    fig, _ = plt.subplots(nrows, ncols, linewidth=2)
        finally:
            mpl.use(backend)

        fig: Figure
        canvas = InteractiveFigureCanvas(fig)
        self.canvas = canvas
        self.figure = fig

        super().__init__()
        self.setLayout(QtW.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(canvas)
        self.setMinimumWidth(180)
        self.setMinimumHeight(135)
        self.resize(180, 135)

        self._editor = QMplEditor()
        self._editor.setParent(self, self._editor.windowFlags())

        canvas.itemPicked.connect(self._edit_artist)
        canvas.clicked.connect(self.as_current_widget)
        canvas.doubleClicked.connect(self._editor.clear)

        self._style = style

    def cla(self) -> None:
        """Clear the current axis."""
        if self._style:
            with plt.style.context(self._style):
                self.ax.cla()
        else:
            self.ax.cla()
        return None

    def draw(self) -> None:
        """Draw (update) the figure."""
        self.figure.tight_layout()
        self.canvas.draw()
        return None

    @property
    def axes(self):
        return self.figure.axes

    @property
    def ax(self) -> Axes:
        """The first matplotlib axis."""
        try:
            _ax = self.axes[0]
        except IndexError:
            _ax = self.figure.add_subplot(111)
        return _ax

    def _reset_canvas(self, fig: Figure, draw: bool = True):
        """Create an interactive figure canvas and add it to the widget."""
        from ._mpl_canvas import InteractiveFigureCanvas

        canvas = InteractiveFigureCanvas(fig)
        self.layout().removeWidget(self.canvas)
        self.canvas = canvas
        self.layout().addWidget(canvas)
        if draw:
            self.draw()

    def _edit_artist(self, artist: Artist):
        from ._artist_editors import pick_container

        cnt = pick_container(artist)
        cnt.changed.connect(self.canvas.draw)
        self._editor.addTab(cnt.native, cnt.get_label())
        self._editor.show()
        return None

    @classmethod
    def current_widget(cls):
        return cls._current_widget

    def as_current_widget(self):
        self.__class__._current_widget = self

    def set_background_color(self, color: str):
        self.figure.set_facecolor(color)
        for ax in self.axes:
            ax.set_facecolor(color)
        self.canvas.draw()


def _use_seaborn_grid(f):
    """
    Some seaborn plot functions will create a new figure.
    This decorator provides a common way to update figure canvas in the widget.
    """

    @wraps(f)
    def func(self: QtMplPlotCanvas, *args, **kwargs):
        backend = mpl.get_backend()
        try:
            mpl.use("Agg")
            grid: Grid = f(self, *args, **kwargs)
        finally:
            mpl.use(backend)

        self._reset_canvas(grid.figure)
        return grid

    return func


class QMplEditor(QtW.QTabWidget):
    """Editor widget for matplotlib artists."""

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setWindowTitle("Matplotlib Artist Editor")

    def addTab(self, widget: QtW.QWidget, label: str) -> int:
        """Add a tab to the editor."""
        area = QtW.QScrollArea(self)
        widget.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        out = super().addTab(area, label)
        area.setWidget(widget)
        return out



from __future__ import annotations
from typing import Any, Hashable, Iterable, Sequence, TYPE_CHECKING
from pandas.core.groupby.generic import DataFrameGroupBy
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Signal, Qt
from collections_undo import arguments

from ._base import QBaseTable, _QTableViewEnhanced, DataFrameModel

if TYPE_CHECKING:
    import pandas as pd


class _QLabeledComboBox(QtW.QWidget):
    currentIndexChanged = Signal(int)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._values: list[Any] = []
        _layout = QtW.QHBoxLayout()
        _layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        _layout.setContentsMargins(2, 2, 2, 2)
        self.setLayout(_layout)
        self._cbox = QtW.QComboBox()
        self._cbox.setMinimumWidth(100)
        self._label = QtW.QLabel()
        _layout.addWidget(self._label)
        _layout.addWidget(self._cbox)

        self._cbox.currentIndexChanged.connect(self.currentIndexChanged.emit)

    def label(self) -> str:
        """Return the label text."""
        return self._label.text()

    def setLabel(self, text: str) -> None:
        """Set the text of label."""
        return self._label.setText(text)

    def choices(self) -> list[Any]:
        return list(map(str, self._values))

    def setChoices(self, items: Iterable[Any]):
        """Set choices of combo box."""
        self._values = list(items)
        self._cbox.clear()
        return self._cbox.addItems(map(str, self._values))

    def currentIndex(self) -> int:
        """Get the current index of the choice."""
        return self._cbox.currentIndex()

    def setCurrentIndex(self, index: int) -> None:
        """Set current index."""
        return self._cbox.setCurrentIndex(index)

    def currentValue(self) -> Any:
        """Current value."""
        return self._values[self.currentIndex()]

    def setCurrentValue(self, value: Any) -> None:
        try:
            index = self._values.index(value)
        except ValueError:
            raise ValueError(f"{value} is not a valid choice.")
        return self.setCurrentIndex(index)

    def copy(self, link: bool = True):
        new = self.__class__()
        new.setLabel(self.label())
        new.setChoices(self.choices())
        new.setCurrentIndex(self.currentIndex())
        new.currentIndexChanged.connect(self.currentIndexChanged.emit)
        if link:
            new.currentIndexChanged.connect(self.setCurrentIndex)
            self.currentIndexChanged.connect(new.setCurrentIndex)

        return new


class _QGroupByWidget(QtW.QWidget):
    _groupby: QTableGroupBy | None = None
    _qtable_view: _QTableViewEnhanced | None = None
    focusedSignal = Signal()

    @classmethod
    def from_groupby(cls, parent: QTableGroupBy) -> _QGroupByWidget:
        self = cls.from_widgets(parent._group_key_cbox, parent._qtable_view)
        self._groupby = parent
        return self

    @classmethod
    def from_widgets(cls, cbox: _QLabeledComboBox, table: _QTableViewEnhanced):
        self = cls()
        _main_layout = QtW.QVBoxLayout()
        _main_layout.setContentsMargins(0, 0, 0, 0)
        _main_layout.addWidget(cbox)
        _main_layout.addWidget(table)
        self.setLayout(_main_layout)
        self._qtable_view = table
        return self

    def copy(self, link: bool = True) -> _QGroupByWidget:
        groupby: QTableGroupBy = self._groupby
        new_cbox = groupby._group_key_cbox.copy(link=link)
        new_table = groupby._qtable_view.copy(link=link)
        new = self.__class__.from_widgets(new_cbox, new_table)
        new_table.setVisible(True)
        new._groupby = groupby
        return new

    @property
    def _selection_model(self):
        return self._groupby._qtable_view._selection_model

    @property
    def _on_moving(self):
        return self._groupby._qtable_view._on_moving

    @property
    def _on_moved(self):
        return self._groupby._qtable_view._on_moved

    def focusInEvent(self, e: QtGui.QFocusEvent) -> None:
        self.focusedSignal.emit()


class QTableGroupBy(QBaseTable):
    _data_raw: DataFrameGroupBy

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        return self._qtable_view_

    @property
    def _central_widget(self) -> _QGroupByWidget:
        return self._central_widget_

    def createQTableView(self):
        self._qtable_view_ = _QTableViewEnhanced(self)
        self._group_key_cbox = _QLabeledComboBox()
        self._group_map: dict[Hashable, Sequence[int]] = {}
        self._group_key_cbox.currentIndexChanged.connect(self._on_cbox_changed)
        wdt = _QGroupByWidget.from_groupby(self)
        self._central_widget_ = wdt
        self.addWidget(wdt)

    def _on_cbox_changed(self):
        return self._set_proxy(self._proxy)

    def getDataFrame(self) -> DataFrameGroupBy:
        return self._data_raw

    @QBaseTable._mgr.interface
    def setDataFrame(self, data: DataFrameGroupBy) -> None:
        if data is None:
            self._data_raw = data
            self._group_key_cbox.setLabel("")

            self._group_map = {}
            self._group_key_cbox.setChoices([])
            self.setProxy(None)
            self._qtable_view.viewport().update()
            return

        if not isinstance(data, DataFrameGroupBy):
            raise TypeError(f"Data must be DataFrameGroupBy, not {type(data)}")
        self._data_raw = data

        # set label
        keys = self._data_raw.keys
        if isinstance(keys, list):
            if len(keys) == 1:
                label = keys[0]
            else:
                label = tuple(keys)
        else:
            label = keys
        self._group_key_cbox.setLabel(f"{label} = ")

        self._group_map = self._data_raw.groups
        self._group_key_cbox.setChoices(self._group_map.keys())
        self.setProxy(None)
        self._qtable_view.viewport().update()
        return

    @setDataFrame.server
    def setDataFrame(self, data) -> None:
        return arguments(getattr(self, "_data_raw", None))

    def createModel(self):
        model = DataFrameModel(self)
        self._qtable_view.setModel(model)
        return None

    def tableSlice(self) -> pd.DataFrame:
        df: pd.DataFrame = self._data_raw.obj
        sl = self._group_map[self._group_key_cbox.currentValue()]
        return df.iloc[sl, :]

    def currentGroup(self) -> Hashable:
        """Return the label of the current group."""
        index = self._group_key_cbox.currentIndex()
        return self._group_key_cbox._values[index]

    def setCurrentGroup(self, group: Hashable) -> None:
        return self._group_key_cbox.setCurrentValue(group)



from __future__ import annotations

from types import MappingProxyType
from typing import Any
from dataclasses import asdict, dataclass, field
from functools import wraps
from pathlib import Path
from contextlib import contextmanager
from appdirs import user_config_dir


TXT_PATH = Path(user_config_dir("tabulous", "tabulous", "history.txt"))
CONFIG_PATH = Path(user_config_dir("tabulous", "tabulous", "config.toml"))
CELL_NAMESPACE_PATH = Path(user_config_dir("tabulous", "tabulous", "cell_namespace.py"))
POST_INIT_PATH = Path(user_config_dir("tabulous", "tabulous", "post_init.py"))


def warn_on_exc(default=None):
    def _wrapper(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                import warnings

                warnings.warn(f"{type(e).__name__}: {e}", UserWarning)
                return default

        return wrapper

    return _wrapper


@warn_on_exc(default=None)
def dump_file_open_path(path: str):
    """Save the recently opened file paths to a text file."""
    path = str(path)
    if not TXT_PATH.exists():
        TXT_PATH.parent.mkdir(parents=True, exist_ok=True)
        TXT_PATH.write_text("")
    with open(TXT_PATH, "r+") as f:
        lines = [s.strip() for s in f.readlines()]
        if path in lines:
            lines.remove(path)
            lines.append(path)
        else:
            lines.append(path)
            if len(lines) > 16:
                lines = lines[-16:]
        f.truncate(0)
        f.seek(0)
        f.write("\n".join(lines))

    return None


@warn_on_exc(default=[])
def load_file_open_path() -> list[str]:
    """Load the recently opened file paths to a text file."""
    if not TXT_PATH.exists():
        return []
    with open(TXT_PATH) as f:
        lines = [s.strip() for s in f.readlines()]

    return lines


def _compile_file(path: Path, default_text: str = ""):
    # if file doesn't exist, create it
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text("")
    with open(path, encoding="utf-8") as f:
        code_str = f.read()
        code = compile(code_str, path, "exec")
    if code_str.strip() == "":
        path.write_text(default_text.strip() + "\n")
    return code


@warn_on_exc(default=MappingProxyType({}))
def load_cell_namespace() -> MappingProxyType:
    """Load the cell namespace from a file."""
    code = _compile_file(CELL_NAMESPACE_PATH, default_text=_CELL_NAMESPACE_TEXT)
    ns: dict[str, Any] = {}
    exec(code, {}, ns)
    to_be_deleted = set()
    for k in ns.keys():
        if k.startswith("_"):
            to_be_deleted.add(k)

    if "__all__" in to_be_deleted:
        _all = ns.get("__all__", [])
        to_be_deleted.update(_all)
    for k in to_be_deleted:
        ns.pop(k, None)
    return MappingProxyType(ns)


def get_post_initializers():
    code = _compile_file(POST_INIT_PATH, default_text=_POST_INIT_TEXT)
    ns: dict[str, Any] = {}
    exec(code, {}, ns)

    if len(ns) == 0:
        # if file is empty, don't do anything
        return None

    from tabulous.post_init import TableInitializer, ViewerInitializer

    table_initializer = TableInitializer()
    viewer_initializer = ViewerInitializer()
    for var in ns.values():
        if isinstance(var, TableInitializer):
            table_initializer.join(var)
        elif isinstance(var, ViewerInitializer):
            viewer_initializer.join(var)
    return viewer_initializer, table_initializer


def prep_default_keybindings() -> dict[str, str | list[str]]:
    from tabulous.commands import DEFAULT_KEYBINDING_SETTING

    kb = {}
    for cmd, seq in DEFAULT_KEYBINDING_SETTING:
        mod = cmd.__module__.split(".")[-1]
        name = cmd.__name__
        kb[f"{mod}.{name}"] = seq
    return kb


@dataclass
class ConsoleNamespace:
    """Default namespace of the console."""

    tabulous: str = "tbl"
    viewer: str = "viewer"
    pandas: str = "pd"
    numpy: str = "np"
    load_startup_file: bool = True


@dataclass
class Table:
    """Table settings."""

    max_row_count: int = 400000
    max_column_count: int = 40000
    font: str = "Arial"
    font_size: int = 10
    row_size: int = 28
    column_size: int = 100


@dataclass
class Cell:
    """Cell settings."""

    eval_prefix: str = "="
    ref_prefix: str = "&="


@dataclass
class Window:
    """Window config."""

    ask_on_close: bool = True
    show_console: bool = False
    theme: str = "light-blue"
    notify_latest: bool = True
    selection_editor: bool = True
    nonmain_style: bool = False
    title_bar: str = "native"


@dataclass
class TabulousConfig:
    """The config model."""

    console_namespace: ConsoleNamespace = field(default_factory=ConsoleNamespace)
    table: Table = field(default_factory=Table)
    cell: Cell = field(default_factory=Cell)
    window: Window = field(default_factory=Window)
    keybindings: dict[str, str | list[str]] = field(
        default_factory=prep_default_keybindings
    )

    @classmethod
    def from_toml(cls, path: Path = CONFIG_PATH) -> TabulousConfig:
        """Load the config file."""
        import toml

        if not path.exists():
            return cls()

        with open(path) as f:
            tm = toml.load(f)
        return cls.from_dict(tm)

    @classmethod
    def from_dict(cls, dict_: dict[str, Any]) -> TabulousConfig:
        console_namespace = dict_.get("console_namespace", {})
        table = dict_.get("table", {})
        cell = dict_.get("cell", {})
        window = dict_.get("window", {})
        if not (kb := dict_.get("keybindings", {})):
            kb = prep_default_keybindings()
        return cls(
            ConsoleNamespace(**_as_fields(console_namespace, ConsoleNamespace)),
            Table(**_as_fields(table, Table)),
            Cell(**_as_fields(cell, Cell)),
            Window(**_as_fields(window, Window)),
            kb,
        )

    def as_toml(self):
        """Save the config file."""
        import toml

        if not CONFIG_PATH.exists():
            CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
            CONFIG_PATH.write_text("")
        with open(CONFIG_PATH, "w") as f:
            toml.dump(asdict(self), f)

        return None

    def as_immutable(self) -> MappingProxyType:
        """Return the immutable version of the config."""
        return MappingProxyType(
            {k: MappingProxyType(v) for k, v in asdict(self).items()}
        )


CONFIG: TabulousConfig | None = None


def get_config(reload: bool = False) -> TabulousConfig:
    """Get the global config."""
    global CONFIG
    if CONFIG is None or reload:
        CONFIG = TabulousConfig.from_toml()
    return CONFIG


def update_config(cfg: TabulousConfig, save: bool = False) -> None:
    """Update the global config."""
    global CONFIG
    CONFIG = cfg
    if save:
        CONFIG.as_toml()
    return None


@contextmanager
def init_config():
    """Temporarily initialize the config."""
    global CONFIG
    if CONFIG is None:
        get_config()
    ori = TabulousConfig()
    old = CONFIG
    CONFIG = ori
    try:
        yield CONFIG
    finally:
        CONFIG = old
        old.as_toml()


def _as_fields(kwargs: dict[str, Any], dcls: type) -> dict[str, Any]:
    """Remove dict keys that does not belong to the dataclass fields."""
    fields = set(dcls.__annotations__.keys())
    return {k: v for k, v in kwargs.items() if k in fields}


_CELL_NAMESPACE_TEXT = """
# File for custom namespace in the table cells.

# Variables defined in the `__all__` list will be available in the cells.
# By, uncommenting following lines, you can use such as
# `=SUM(df.iloc[:, 0])` in cells.

# import scipy
# import numpy as np

# __all__ = ["scipy", "SUM", "AVERAGE"]

# def SUM(x):
#     return np.sum(x)

# def AVERAGE(x):
#     return np.mean(x)
"""

_POST_INIT_TEXT = """
# File for post-initialization of the viewer and table.
# You can add custom actions, commands and variables on the startup of the
# application.

# First, get initializer objects.

# from tabulous.post_init import get_initializers
# viewer, table = get_initializers()

# These objects has similar interface to the viewer and table classes.

# 1. Add custom actions to the right-click context menu of table columns.

# @table.columns.register("User defined > Print column name")
# def _print_column_name(table, index):
#     print(table.columns[index])

# 2. Add custom keybindings to viewers and tables.

# @viewer.keymap.register("Ctrl+K, Ctrl+1")
# def _my_keybinding(viewer):
#     print(viewer)

# @table.keymap.register("Ctrl+Shift+K, Ctrl+1")
# def _my_keybinding(table):
#     print(table)

# 3. Add custom variables to the console (note that you can also update
#    the IPython startup files to do the same thing).

# viewer.console.update({"PI": 3.141592653589793})

# 4. Add custom variables to the table cells (not that you can also update
#    the cell_namespace.py file to do the same thing).

# import numpy as np
# viewer.cell_namespace.update({"SUM": np.sum})

"""



from setuptools import setup, find_packages

TABULOUS = "tabulous"

with open(f"{TABULOUS}/__init__.py", encoding="utf-8") as f:
    line = next(iter(f))
    VERSION = line.strip().split()[-1][1:-1]

with open("README.md") as f:
    README = f.read()

setup(
    name=TABULOUS,
    version=VERSION,
    description="A table data viewer for Python",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Hanjin Liu",
    author_email="liuhanjin-sc@g.ecc.u-tokyo.ac.jp",
    license="BSD 3-Clause",
    download_url="https://github.com/hanjinliu/tabulous",
    packages=find_packages(exclude=["docs", "examples", "rst", "tests", "tests.*"]),
    package_data={
        TABULOUS: ["**/*.pyi", "*.pyi", "**/*.svg", "**/*.png", "**/*.qss", "**/*.json"]
    },
    include_package_data=True,
    install_requires=[
        "magicgui>=0.5.1",
        "psygnal>=0.6.1",
        "qtpy>=1.10.0",
        "pandas>=1.5.2",
        "collections-undo>=0.0.7",
        "appdirs>=1.4.4",
        "qtconsole",
        "qt-command-palette>=0.0.5",
        "toml",
        "matplotlib>=3.1",
        "tabulate",
        "requests",
    ],
    extras_require={
        "all": [
            "seaborn>=0.11",
            "pyqt5>=5.12.3",
            "scipy>=1.7",
            "scikit-learn>=1.1",
        ],
        "pyqt5": ["pyqt5>=5.12.3"],
        "pyqt6": ["pyqt6>=6.3.1"],
        "scikit-learn": ["scikit-learn>=1.0"],
    },
    entry_points={
        "console_scripts": [f"{TABULOUS}={TABULOUS}.__main__:main"],
    },
    python_requires=">=3.8",
)



from __future__ import annotations
from typing import Generic, Sequence, TypeVar
import weakref
from matplotlib.collections import PathCollection, LineCollection, PolyCollection
from matplotlib.container import BarContainer
from matplotlib.lines import Line2D
from matplotlib.text import Text
from matplotlib.patches import Polygon

from matplotlib.artist import Artist

from magicgui.widgets import (
    ComboBox,
    Container,
    SpinBox,
    FloatSpinBox,
    LineEdit,
    Widget,
)

from ._artist_types import (
    LineStyle,
    Marker,
    HatchStyle,
    VerticalAlignment,
    HorizontalAlignment,
)

_A = TypeVar("_A", bound=Artist)
_T = TypeVar("_T")

_TYPE_MAP: dict[type[Artist], type[ArtistEditor]] = {}


def _register(artist: type[Artist]):
    def _wrapper(cls: _T) -> _T:
        _TYPE_MAP[artist] = cls
        return cls

    return _wrapper


class _Ref(Generic[_T]):
    """a non-weak reference type."""

    def __init__(self, seq: _T):
        self._seq = seq

    def __call__(self) -> _T:
        return self._seq


def _to_float_rgba(rgba: Sequence[int]) -> list[float]:
    return [c / 255 for c in rgba]


class ArtistEditor(Container, Generic[_A]):
    """Base class for artist editor containers."""

    def __init__(self, artist: _A):
        self.artist = artist
        widgets = self._create_widgets()
        super().__init__(widgets=widgets)

    def _create_widgets(self) -> list[Widget]:
        raise NotImplementedError()

    def get_label(self):
        """Get the label for the artist."""
        return self.artist.get_label()

    @property
    def artist(self) -> _A:
        """Return the Artist object."""
        out = self._artist()
        if out is None:
            raise ValueError("Artist object has been deleted.")
        return out

    @artist.setter
    def artist(self, artist: _A):
        if isinstance(artist, (tuple, list)):  # Collection type is a subclass of tuple
            self._artist = _Ref(artist)
        else:
            self._artist = weakref.ref(artist)
        return None


@_register(Line2D)
class Line2DEdit(ArtistEditor[Line2D]):
    def _create_widgets(self) -> None:
        from tabulous._magicgui import ColorEdit

        widgets = []
        line = self.artist

        # line color
        color = line.get_color()
        if not isinstance(color, str):
            color = [int(c * 255) for c in color]
        _color_edit = ColorEdit(name="color", value=color)
        _color_edit.changed.connect(self.set_color)
        widgets.append(_color_edit)

        # line style
        _ls_edit = ComboBox(
            choices=LineStyle, value=LineStyle(line.get_linestyle()), label="linestyle"
        )
        _ls_edit.changed.connect(self.set_linestyle)
        widgets.append(_ls_edit)

        # line width
        _lw_edit = FloatSpinBox(
            min=0.0, max=10.0, step=0.5, value=line.get_linewidth(), label="linewidth"
        )
        _lw_edit.changed.connect(self.set_linewidth)
        widgets.append(_lw_edit)

        _marker = ComboBox(
            choices=Marker, label="marker", value=Marker(line.get_marker())
        )
        _marker.changed.connect(self.set_marker)
        _markerfacecolor = ColorEdit(
            label="marker face color", value=fix_color(line.get_markerfacecolor())
        )
        _markerfacecolor.changed.connect(self.set_markerfacecolor)
        _markeredgecolor = ColorEdit(
            label="marker edge color", value=fix_color(line.get_markeredgecolor())
        )
        _markeredgecolor.changed.connect(self.set_markeredgecolor)
        _markeredgewidth = FloatSpinBox(
            min=0.0,
            max=10.0,
            step=0.5,
            label="marker edge width",
            value=line.get_markeredgewidth(),
        )
        _markeredgewidth.changed.connect(self.set_markeredgewidth)
        _markersize = FloatSpinBox(
            label="marker size",
            min=0.0,
            max=50.0,
            step=0.5,
            value=line.get_markersize(),
        )
        _markersize.changed.connect(self.set_markersize)

        self._marker_related: list[Widget] = [
            _markerfacecolor,
            _markeredgecolor,
            _markeredgewidth,
            _markersize,
        ]

        widgets.extend(
            [_marker, _markerfacecolor, _markeredgecolor, _markeredgewidth, _markersize]
        )

        # zorder
        _zorder = SpinBox(min=-10000, max=10000, value=line.get_zorder(), name="zorder")
        _zorder.changed.connect(self.set_zorder)
        widgets.append(_zorder)

        self.set_marker(_marker.value)
        return widgets

    def set_color(self, rgba: tuple[int, int, int, int]) -> None:
        """Set the line color."""
        self.artist.set_color(_to_float_rgba(rgba))

    def set_linestyle(self, ls: LineStyle):
        self.artist.set_linestyle(ls.value)

    def set_linewidth(self, lw: float):
        self.artist.set_linewidth(lw)

    def set_zorder(self, zorder: int):
        self.artist.set_zorder(zorder)

    def set_marker(self, marker: Marker):
        marker = Marker(marker)
        self.artist.set_marker(marker.value)
        has_marker = marker != Marker.none
        for wdt in self._marker_related:
            wdt.enabled = has_marker

    def set_markerfacecolor(self, rgba):
        self.artist.set_markerfacecolor(_to_float_rgba(rgba))

    def set_markeredgecolor(self, rgba):
        self.artist.set_markeredgecolor(_to_float_rgba(rgba))

    def set_markeredgewidth(self, width: float):
        self.artist.set_markeredgewidth(width)

    def set_markersize(self, size: float):
        self.artist.set_markersize(size)


@_register(PathCollection)
class ScatterEdit(ArtistEditor[PathCollection]):
    def _create_widgets(self) -> list[Widget]:
        from tabulous._magicgui._color_edit import ColorEdit

        scatter = self.artist

        # scatter color
        _facecolor = ColorEdit(
            name="face color", value=scatter.get_facecolor()[0] * 255
        )
        _facecolor.changed.connect(self.set_facecolor)

        _edgecolor = ColorEdit(
            name="edge color", value=scatter.get_edgecolor()[0] * 255
        )
        _edgecolor.changed.connect(self.set_edgecolor)

        # _edgewidth = FloatSpinBox(name="edge width", value=scatter.get_linewidth())

        # marker
        _marker_edit = ComboBox(choices=Marker, value=Marker.circle, name="marker")
        _marker_edit.changed.connect(self.set_marker)

        _size_edit = FloatSpinBox(
            min=0.0, max=500.0, step=1, value=scatter.get_sizes()[0], name="size"
        )
        _size_edit.changed.connect(self.set_size)

        # zorder
        _zorder = SpinBox(
            min=-10000, max=10000, value=scatter.get_zorder(), name="zorder"
        )
        _zorder.changed.connect(self.set_zorder)

        return [_facecolor, _edgecolor, _marker_edit, _size_edit, _zorder]

    def set_facecolor(self, rgba: tuple[int, int, int, int]) -> None:
        """Set face colors of the scatter."""
        self.artist.set_facecolor(_to_float_rgba(rgba))

    def set_edgecolor(self, rgba: tuple[int, int, int, int]) -> None:
        """Set edge colors of the scatter."""
        self.artist.set_edgecolors(_to_float_rgba(rgba))

    def set_marker(self, marker: Marker | str):
        import matplotlib.markers as mmarkers

        marker = Marker(marker)
        marker_obj = mmarkers.MarkerStyle(marker.value)

        path = marker_obj.get_path().transformed(marker_obj.get_transform())
        self.artist.set_paths((path,))

    def set_size(self, size: int) -> None:
        self.artist.set_sizes([size])

    def set_zorder(self, zorder: int):
        self.artist.set_zorder(zorder)


@_register(BarContainer)
class PatchContainerEdit(ArtistEditor[BarContainer]):
    def _create_widgets(self) -> list[Widget]:
        from ..._magicgui._color_edit import ColorEdit

        artist = self.artist
        _facecolor = ColorEdit(
            label="face color", value=fix_color(artist[0].get_facecolor())
        )
        _facecolor.changed.connect(self.set_facecolor)
        _edgecolor = ColorEdit(
            label="edge color", value=fix_color(artist[0].get_edgecolor())
        )
        _edgecolor.changed.connect(self.set_edgecolor)
        hatch = artist[0].get_hatch() or "None"
        _hatch = ComboBox(choices=HatchStyle, label="hatch", value=HatchStyle(hatch))
        _hatch.changed.connect(self.set_hatch)
        return [_facecolor, _edgecolor, _hatch]

    def set_facecolor(self, rgba):
        rgba = _to_float_rgba(rgba)
        for artist in self.artist:
            artist.set_facecolor(rgba)

    def set_edgecolor(self, rgba):
        rgba = _to_float_rgba(rgba)
        for artist in self.artist:
            artist.set_edgecolor(rgba)

    def set_hatch(self, hatch: HatchStyle):
        if hatch == HatchStyle.none:
            v = None
        else:
            v = hatch.value
        for artist in self.artist:
            artist.set_hatch(v)

    def get_label(self) -> str:
        return self.artist[0].get_label()


@_register(LineCollection)
class LineCollectionEdit(ArtistEditor[LineCollection]):
    def _create_widgets(self) -> list[Widget]:
        from tabulous._magicgui._color_edit import ColorEdit

        line = self.artist

        _color = ColorEdit(name="color", value=fix_color(line.get_color()[0]))
        _color.changed.connect(self.set_errorbar_color)

        # line width
        _lw_edit = FloatSpinBox(
            min=0.0, max=10.0, step=0.5, value=line.get_linewidth(), name="linewidth"
        )
        _lw_edit.changed.connect(self.set_linewidth)

        return [_color, _lw_edit]

    def set_errorbar_color(self, rgba):
        self.artist.set_color(_to_float_rgba(rgba))

    def set_linewidth(self, width: float):
        self.artist.set_linewidth(width)


@_register(PolyCollection)
class PolyCollectionEdit(ArtistEditor[PolyCollection]):
    def _create_widgets(self) -> list[Widget]:
        from tabulous._magicgui._color_edit import ColorEdit

        poly = self.artist

        ec = poly.get_edgecolors()
        if ec.shape[0] == 0:
            ec = (0, 0, 0, 0)
        else:
            ec = ec[0]

        _face_color = ColorEdit(
            name="facecolor", value=fix_color(poly.get_facecolors()[0])
        )
        _face_color.changed.connect(self.set_facecolor)
        _edge_color = ColorEdit(name="edgecolor", value=fix_color(ec))
        _edge_color.changed.connect(self.set_edgecolor)
        _hatch = ComboBox(choices=HatchStyle, name="hatch", value=HatchStyle.none)
        _hatch.changed.connect(self.set_hatch)
        _ls = ComboBox(choices=LineStyle, name="linestyle", value=LineStyle.solid)
        _ls.changed.connect(self.set_linestyle)

        # line width
        _lw_edit = FloatSpinBox(
            min=0.0, max=10.0, step=0.5, value=poly.get_linewidth(), name="linewidth"
        )
        _lw_edit.changed.connect(self.artist.set_linewidth)

        return [_face_color, _edge_color, _ls, _lw_edit]

    def set_facecolor(self, rgba):
        self.artist.set_facecolor(_to_float_rgba(rgba))

    def set_edgecolor(self, rgba):
        self.artist.set_edgecolor(_to_float_rgba(rgba))

    def set_hatch(self, hatch: HatchStyle):
        if hatch == HatchStyle.none:
            v = None
        else:
            v = hatch.value
        self.artist.set_hatch(v)

    def set_linestyle(self, ls: LineStyle):
        self.artist.set_linestyle(ls.value)


@_register(Polygon)
class PolygonEdit(ArtistEditor[Polygon]):
    def _create_widgets(self) -> list[Widget]:
        from tabulous._magicgui._color_edit import ColorEdit

        polygon = self.artist
        _facecolor = ColorEdit(
            label="face color", value=fix_color(polygon.get_facecolor())
        )
        _facecolor.changed.connect(self.set_facecolor)
        _edgecolor = ColorEdit(
            label="edge color", value=fix_color(polygon.get_edgecolor())
        )
        _edgecolor.changed.connect(self.set_edgecolor)

        _linewidth = FloatSpinBox(
            label="edge width",
            value=polygon.get_linewidth(),
            min=0.0,
            max=50.0,
            step=0.5,
        )
        _linewidth.changed.connect(self.set_linewidth)

        return [_facecolor, _edgecolor, _linewidth]

    def set_facecolor(self, rgba):
        rgba = _to_float_rgba(rgba)
        self.artist.set_facecolor(rgba)

    def set_edgecolor(self, rgba):
        rgba = _to_float_rgba(rgba)
        self.artist.set_edgecolor(rgba)

    def set_linewidth(self, width: float):
        self.artist.set_linewidth(width)


@_register(Text)
class TextEdit(ArtistEditor[Text]):
    def _create_widgets(self) -> list[Widget]:
        from tabulous._magicgui._color_edit import ColorEdit

        _color = ColorEdit(name="color", value=fix_color(self.artist.get_color()))
        _color.changed.connect(self.set_color)

        _font_size = SpinBox(
            name="font size", value=self.artist.get_fontsize(), min=1, max=180
        )
        _font_size.changed.connect(self.artist.set_fontsize)

        _text = LineEdit(name="text", value=self.artist.get_text())
        _text.changed.connect(self.artist.set_text)

        _va = ComboBox(
            name="vertical",
            choices=VerticalAlignment,
            value=VerticalAlignment(self.artist.get_verticalalignment()),
        )
        _ha = ComboBox(
            name="horizontal",
            choices=HorizontalAlignment,
            value=HorizontalAlignment(self.artist.get_horizontalalignment()),
        )
        _va.changed.connect(self.set_verticalalignment)
        _ha.changed.connect(self.set_horizontalalignment)

        _alignment = Container(widgets=[_va, _ha], name="alignment")
        return [_color, _font_size, _text, _alignment]

    def set_color(self, rgba):
        self.artist.set_color(_to_float_rgba(rgba))

    def set_verticalalignment(self, alignment: VerticalAlignment):
        self.artist.set_verticalalignment(alignment.value)

    def set_horizontalalignment(self, alignment: HorizontalAlignment):
        self.artist.set_horizontalalignment(alignment.value)


def pick_container(artist: Artist) -> ArtistEditor:
    """Return a proper container for the given artist."""
    for artist_cls, wdt_cls in _TYPE_MAP.items():
        if isinstance(artist, artist_cls):
            return wdt_cls(artist)
    raise ValueError(f"No container found for artist {type(artist).__name__}.")


def fix_color(color):
    if not isinstance(color, str):
        color = [int(c * 255) for c in color]
    return color



from ._table import QTableLayer
from ._spreadsheet import QSpreadSheet
from ._groupby import QTableGroupBy
from ._display import QTableDisplay
from ._base import QMutableTable, QBaseTable, QTableGroup

__all__ = [
    "QTableLayer",
    "QSpreadSheet",
    "QBaseTable",
    "QMutableTable",
    "QTableGroupBy",
    "QTableDisplay",
    "QTableGroup",
]



from __future__ import annotations

import re
from typing import Any, TYPE_CHECKING, Callable, overload, TypeVar, Generic
from typing_extensions import ParamSpec

if TYPE_CHECKING:
    from tabulous._qt._action_registry import QActionRegistry

_S = TypeVar("_S")
_T = TypeVar("_T")
_P = ParamSpec("_P")


class SupportActionRegistration(Generic[_S, _T]):
    """An abstract class to support action registration to right-click contextmenu."""

    def _get_qregistry(self) -> QActionRegistry:
        raise NotImplementedError()

    @property
    def parent(self) -> _S:
        raise NotImplementedError()

    # fmt: off
    @overload
    def register(self, location: str) -> Callable[[Callable[[_S, _T], Any]], Callable[[_S, _T], Any]]: ...  # noqa: E501
    @overload
    def register(self, location: str, func: Callable[[_S, _T], Any]) -> Callable[[_S, _T], Any]: ...  # noqa: E501
    @overload
    def register(self, func: Callable[[_S, _T], Any]) -> Callable[[_S, _T], Any]: ...  # noqa: E501
    # fmt: on

    def register(self, *args):
        """Register an contextmenu action."""
        reg = self._get_qregistry()
        nargs = len(args)
        if nargs == 0 or nargs > 2:
            raise TypeError("One or two arguments are allowed.")
        if nargs == 1:
            arg = args[0]
            if callable(arg):
                loc, func = getattr(arg, "__name__", repr(arg)), arg
            else:
                loc, func = arg, None
        else:
            loc, func = args

        # check type
        if not isinstance(loc, str) or (func is not None and not callable(func)):
            arg = type(loc).__name__
            if func is not None:
                arg += f", {type(func).__name__}"
            raise TypeError(
                f"No overloaded method matched the input ({arg}).\n"
                "1. register(location: str)\n"
                "2. register(func: Callable)\n"
                "3. register(location: str, func: Callable)"
            )

        def wrapper(f: Callable[[_S, int], Any]):
            meth = _NormalizedFunction(f, self.parent)
            reg.registerAction(loc, meth)
            return f

        return wrapper if func is None else wrapper(func)

    def unregister(self, location: str):
        """Unregister an contextmenu action."""
        self._get_qregistry().unregisterAction(location)


def register_action(self: SupportActionRegistration, *args):
    """Register an contextmenu action."""
    import warnings

    warnings.warn(
        "`register_action` is deprecated. Use `register` instead.",
        DeprecationWarning,
    )

    reg = self._get_qregistry()
    nargs = len(args)
    if nargs == 0 or nargs > 2:
        raise TypeError("One or two arguments are allowed.")
    if nargs == 1:
        arg = args[0]
        if callable(arg):
            loc, func = getattr(arg, "__name__", repr(arg)), arg
        else:
            loc, func = arg, None
    else:
        loc, func = args

    # check type
    if not isinstance(loc, str) or (func is not None and not callable(func)):
        arg = type(loc).__name__
        if func is not None:
            arg += f", {type(func).__name__}"
        raise TypeError(f"No overloaded method matched the input ({arg}).")

    def wrapper(f: Callable[[int], Any]):
        reg.registerAction(loc, f)
        return f

    return wrapper if func is None else wrapper(func)


SupportActionRegistration.register_action = register_action

# e.g. f() takes from 0 to 1 positional arguments but 2 were given
_PATTERN = re.compile(r".*takes .* positional arguments? but (\d+) w.+ given")


class _NormalizedFunction(Generic[_P]):
    def __init__(self, f: Callable[_P, Any], parent) -> None:
        self._f = f
        self._parent = parent
        self._f_normed = f

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} of {self._f.__name__!r}>"

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs):
        try:
            out = self._f_normed(self._parent, *args, **kwargs)
        except TypeError as e:
            if _PATTERN.match(str(e)):
                import inspect

                sig = inspect.signature(self._f)
                nparams = len(sig.parameters)

                if nparams == 0:

                    def normed(parent, *args, **kwargs):
                        return self._f(**kwargs)

                else:

                    def normed(parent, *args, **kwargs):
                        return self._f(parent, *args[: nparams - 1], **kwargs)

                self._f_normed = normed
                out = normed(self._parent, *args)
            else:
                raise e
        return out

    @property
    def func(self):
        return self._f



from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from functools import partial
from abc import ABC, abstractmethod
import re
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Signal, Qt

from . import _utils

if TYPE_CHECKING:
    from .._table import QBaseTable


class MatchMode:
    """Match mode for search"""

    value = "12"
    text = "'12'"
    text_partial = "'1'2"
    regex = ".*"
    expr = ">>>"


class SearchOrientation:
    """Orientation of searching"""

    row = "⇉"
    column = "⇊"


class QComboButtons(QtW.QWidget):
    """Collection of buttons to choose between match modes"""

    currentTextChanged = Signal(str)

    def __init__(self) -> None:
        super().__init__()
        choices = list(
            v for k, v in MatchMode.__dict__.items() if not k.startswith("_")
        )
        _layout = QtW.QHBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        _layout.setSpacing(2)
        self.setLayout(_layout)
        buttons = [QtW.QPushButton(choice, self) for choice in choices]
        buttons[0].setToolTip("Match by value")
        buttons[1].setToolTip("Match by text")
        buttons[2].setToolTip("Partial match by text")
        buttons[3].setToolTip("Match by regular expression")
        buttons[4].setToolTip("Match by any expression")

        for i, btn in enumerate(buttons):
            _layout.addWidget(btn)
            btn.setCheckable(True)
            btn.setFont(QtGui.QFont("Arial"))
            btn.setFixedSize(32, 18)
            btn.clicked.connect(lambda checked, idx=i: self.setCurrentIndex(idx))

        self._buttons = buttons
        self.setCurrentIndex(0)

    def currentIndex(self) -> int:
        """Currently selected index."""
        for i, btn in enumerate(self._buttons):
            if btn.isChecked():
                return i
        return -1

    def setCurrentIndex(self, index: int) -> None:
        """Set currently selected index."""
        for i, btn in enumerate(self._buttons):
            checked = i == index
            btn.setChecked(checked)
            font = btn.font()
            font.setBold(checked)
            btn.setFont(font)
        text = self._buttons[index].text()
        self.currentTextChanged.emit(text)


class QFinderWidget(QtW.QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        _layout = QtW.QVBoxLayout()
        self._search_box = QSearchBox()
        self._search_box.enterClicked.connect(self.findNext)
        self._search_box.textChanged.connect(self.initSearchBox)
        _search_box_widget = QWithButtons(self._search_box, texts=["↑", "↓"])
        _search_box_widget.clicked.connect(
            lambda i: self.findPrevious() if i == 0 else self.findNext()
        )
        _search_box_widget.button(0).setToolTip("Find next")
        _search_box_widget.button(1).setToolTip("Find previous")
        _layout.addWidget(_search_box_widget)

        self._replace_box = QSearchBox()
        self._replace_box.enterClicked.connect(self.replaceCurrent)
        _replace_box_widget = QWithButtons(self._replace_box, texts=["↵", "↵↵"])
        _replace_box_widget.button(0).setToolTip("Replace current item")
        _replace_box_widget.button(1).setToolTip("Replace all items")
        _layout.addWidget(_replace_box_widget)
        _replace_box_widget.clicked.connect(
            lambda i: self.replaceCurrent() if i == 0 else self.replaceAll()
        )

        policy = (QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed)
        self._search_box.setSizePolicy(*policy)
        self._replace_box.setSizePolicy(*policy)

        _footer = QtW.QWidget()
        _layout.addWidget(_footer)
        _footer.setLayout(QtW.QHBoxLayout())
        _footer.layout().setContentsMargins(0, 0, 0, 0)

        self.cbox_match = QComboButtons()
        self.cbox_match.currentTextChanged.connect(self.setMatchMode)

        self.cbox_ori = QtW.QComboBox()
        self.cbox_ori.addItems([SearchOrientation.row, SearchOrientation.column])
        self.cbox_ori.setCurrentIndex(1)
        self.cbox_ori.currentTextChanged.connect(self.initSearchBox)

        _footer.layout().addWidget(self.cbox_ori)
        _footer.layout().addWidget(self.cbox_match)

        self.setLayout(_layout)
        self._qtable_viewer = _utils.find_parent_table_viewer(self)
        self._current_iterator: TwoWayIterator | None = None

        self.setMatchMode(MatchMode.value)
        self._current_index = None

    def searchBox(self) -> QSearchBox:
        return self._search_box

    def replaceBox(self) -> QSearchBox:
        return self._replace_box

    def setReplaceBoxVisible(self, visible: bool):
        return self._replace_box.parentWidget().setVisible(visible)

    def initSearchBox(self):
        if self.cbox_ori.currentText() == SearchOrientation.row:
            self._current_iterator = RowwiseIterator(self.currentTable().dataShape())
        else:
            self._current_iterator = ColumnwiseIterator(self.currentTable().dataShape())
        if self._current_index is not None:
            self._current_iterator.set_index(*self._current_index)

    def findNext(self) -> None:
        """Find next item that match in the current mode."""
        text = self._search_box.text()
        if not text:
            return
        qtable = self.currentTable()
        pf = partial(self._match_method, qtable, text)
        self._current_iterator.shape = qtable.dataShape()

        try:
            r, c = self._current_iterator.next_until(pf)
        except ItemNotFound:
            raise ItemNotFound(f"{text!r} not found.")
        qtable = self.currentTable()
        qtable.moveToItem(r + 2, c + 2)
        qtable.moveToItem(r, c)
        qtable.setSelections([(r, c)])
        self._current_index = (r, c)
        return

    def findPrevious(self) -> None:
        """Find previous item that match in the current mode."""
        text = self._search_box.text()
        if not text:
            return
        qtable = self.currentTable()
        pf = partial(self._match_method, qtable, text)
        self._current_iterator.shape = qtable.dataShape()

        try:
            r, c = self._current_iterator.prev_until(pf)
        except ItemNotFound:
            raise ItemNotFound(f"{text!r} not found.")
        qtable = self.currentTable()
        qtable.moveToItem(r + 2, c + 2)
        qtable.moveToItem(r, c)
        qtable.setSelections([(r, c)])
        self._current_index = (r, c)
        return

    def replaceCurrent(self) -> None:
        """Replace the current cell with the text in the box"""
        if self._current_index is None or not self._replace_box.isVisible():
            return
        qtable = self.currentTable()
        r, c = self._current_index
        convert_value = qtable._get_converter(c)
        value = convert_value(c, self._replace_box.text())
        qtable.setDataFrameValue(r, c, value)
        return self.findNext()

    def replaceAll(self) -> None:
        """Replace all matching cells with the text in the box"""
        text = self._search_box.text()
        text_after = self._replace_box.text()
        if not text:
            return
        self.initSearchBox()
        qtable = self.currentTable()
        pf = partial(self._match_method, qtable, text)
        self._current_iterator.shape = qtable.dataShape()

        with qtable._mgr.merging(lambda cmd: f"Replace {text!r} to {text_after!r}"):
            while True:
                try:
                    r, c = self._current_iterator.next_until(pf)
                except ItemNotFound:
                    self.initSearchBox()
                    break
                convert_value = qtable._get_converter(c)
                value = convert_value(c, text_after)
                qtable.setDataFrameValue(r, c, value)

        return None

    def setMatchMode(self, mode: str):
        """Set the match mode and update match function."""
        if mode == MatchMode.value:
            self._match_method = self._value_match
        elif mode == MatchMode.text:
            self._match_method = self._text_match
        elif mode == MatchMode.text_partial:
            self._match_method = self._text_partial_match
        elif mode == MatchMode.regex:
            self._match_method = self._text_regex_match
        elif mode == MatchMode.expr:
            self._match_method = self._expr_match
        else:
            raise ValueError(f"Unknown match mode: {mode}")

    def currentTable(self) -> QBaseTable:
        tablestack = self._qtable_viewer._tablestack
        idx = tablestack.currentIndex()
        return tablestack.tableAtIndex(idx)

    def _value_match(self, qtable: QBaseTable, text: str, r: int, c: int) -> bool:
        try:
            val = qtable.convertValue(c, text)
        except Exception:
            return False
        return qtable.dataShown().iloc[r, c] == val

    def _text_match(self, qtable: QBaseTable, text: str, r: int, c: int) -> bool:
        model = qtable.model()
        index = model.index(r, c)
        displayed_text = qtable.model().data(index, Qt.ItemDataRole.DisplayRole)
        return displayed_text == text

    def _text_partial_match(
        self, qtable: QBaseTable, text: str, r: int, c: int
    ) -> bool:
        return text in str(qtable.dataShown().iloc[r, c])

    def _text_regex_match(self, qtable: QBaseTable, text: str, r: int, c: int) -> bool:
        return re.match(text, str(qtable.dataShown().iloc[r, c])) is not None

    def _expr_match(self, qtable: QBaseTable, text: str, r: int, c: int) -> bool:
        import numpy, pandas

        f = eval(f"(lambda x: {text})", {"np": numpy, "pd": pandas}, {})
        try:
            return f(qtable.dataShown().iloc[r, c])
        except Exception:
            return False


class QSearchBox(QtW.QLineEdit):
    enterClicked = Signal()
    escClicked = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        from tabulous._utils import get_config

        table_config = get_config().table
        self.setFont(QtGui.QFont(table_config.font, table_config.font_size))

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() == Qt.Key.Key_Return:
            self.enterClicked.emit()
        elif event.key() == Qt.Key.Key_Escape:
            self.escClicked.emit()
        else:
            super().keyPressEvent(event)


class QWithButtons(QtW.QWidget):
    """A container widget that contains a line edit and a button."""

    clicked = Signal(int)

    def __init__(self, widget: QtW.QLineEdit, texts: list[str]) -> None:
        super().__init__()
        _layout = QtW.QHBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        _layout.addWidget(widget)
        self._buttons = []
        for i, text in enumerate(texts):
            btn = self._make_button(i, text)
            _layout.addWidget(btn)
            self._buttons.append(btn)

        self.setLayout(_layout)

    def _make_button(self, i: int, text: str):
        btn = QtW.QPushButton(text)
        btn.clicked.connect(lambda: self.clicked.emit(i))
        btn.setFixedWidth(36)
        return btn

    def button(self, idx: int) -> QtW.QPushButton:
        """Get the button at the given index."""
        return self._buttons[idx]


class ItemNotFound(Exception):
    """Raised when the item is not found."""


class TwoWayIterator(ABC):
    def __init__(self, shape: tuple[int, int]):
        self.shape = shape
        self.go_to_first()

    def set_index(self, r: int, c: int) -> None:
        self._r = r
        self._c = c
        return None

    @abstractmethod
    def go_to_first(self) -> None:
        """Go to the first index."""

    @abstractmethod
    def go_to_last(self) -> None:
        """Go to the last index."""

    @abstractmethod
    def next(self):
        """Next item."""

    @abstractmethod
    def prev(self):
        """Previous item."""

    def next_until(self, predicate: Callable[[int, int], bool]) -> tuple[int, int]:
        """Next item until the predicate is True."""
        start = item = self.next()
        while True:
            if predicate(*item):
                return item
            item = self.next()
            if item == start:
                raise ItemNotFound

    def prev_until(self, predicate: Callable[[int, int], bool]) -> tuple[int, int]:
        """Next item until the predicate is True."""
        start = item = self.prev()
        while True:
            if predicate(*item):
                return item
            item = self.prev()
            if item == start:
                raise ItemNotFound


class RowwiseIterator(TwoWayIterator):
    def go_to_first(self) -> None:
        self._r = 0
        self._c = -1

    def go_to_last(self) -> None:
        self._r = self.shape[0] - 1
        self._c = self.shape[1]

    def next(self) -> tuple[int, int]:
        self._c += 1
        nr, nc = self.shape
        if self._c >= nc:
            self._c = 0
            self._r += 1
            if self._r >= nr:
                self.go_to_first()
                self._c += 1
        return self._r, self._c

    def prev(self) -> tuple[int, int]:
        self._c -= 1
        if self._c < 0:
            nr, nc = self.shape
            self._c = nc - 1
            self._r -= 1
            if self._r < 0:
                self.go_to_last()
                self._c -= 1
        return self._r, self._c


class ColumnwiseIterator(TwoWayIterator):
    def go_to_first(self) -> None:
        self._r = -1
        self._c = 0

    def go_to_last(self) -> None:
        self._r = self.shape[0]
        self._c = self.shape[1] - 1

    def next(self) -> tuple[int, int]:
        self._r += 1
        nr, nc = self.shape
        if self._r >= nr:
            self._r = 0
            self._c += 1
            if self._c >= nc:
                self.go_to_first()
                self._r += 1
        return self._r, self._c

    def prev(self) -> tuple[int, int]:
        self._r -= 1
        if self._r < 0:
            nr, nc = self.shape
            self._r = nr - 1
            self._c -= 1
            if self._c < 0:
                self.go_to_last()
                self._r -= 1
        return self._r, self._c



from __future__ import annotations
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, Signal
from tabulous._qt._clickable_label import QClickableLabel
from tabulous._utils import load_file_open_path
from tabulous import commands as cmds

if TYPE_CHECKING:
    from .._mainwindow._base import _QtMainWidgetBase

_HEIGHT = 24


class QStartupWidget(QtW.QWidget):
    """The startup widget which is shown when no table exists."""

    def __init__(self, parent=None):
        super().__init__(parent)
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(12, 12, 12, 12)
        _layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(_layout)

        # fmt: off
        self._open_table_btn = QClickableLabel("Open File as Table (Ctrl+O)")  # noqa: E501
        self._open_spreadsheet_btn = QClickableLabel("Open File as Spreadsheet (Ctrl+K, Ctrl+O)")  # noqa: E501
        self._open_new_btn = QClickableLabel("New Spreadsheet (Ctrl+N)")  # noqa: E501
        self._path_list = QPathList()

        self._open_table_btn.clicked.connect(lambda: cmds.file.open_table(self.mainWidget()._table_viewer))  # noqa: E501
        self._open_spreadsheet_btn.clicked.connect(lambda: cmds.file.open_spreadsheet(self.mainWidget()._table_viewer))  # noqa: E501
        self._open_new_btn.clicked.connect(lambda: cmds.table.new_spreadsheet(self.mainWidget()._table_viewer))  # noqa: E501
        self._path_list.pathClicked.connect(lambda path: self.mainWidget()._table_viewer.open(path))  # noqa: E501
        # fmt: on
        _layout.addWidget(self._open_table_btn)
        _layout.addWidget(self._open_spreadsheet_btn)
        _layout.addWidget(self._open_new_btn)
        _layout.addWidget(self._path_list)
        self.setMinimumSize(0, 0)
        return None

    def mainWidget(self) -> _QtMainWidgetBase:
        """The parent main widget."""
        return self.parent().parent().parent()

    def widget(self, i):
        return self

    def resizeEvent(self, a0: QtGui.QResizeEvent) -> None:
        h = a0.size().height()
        rest = max(h - _HEIGHT * 2 - 30, 0)
        n_show = rest // (_HEIGHT + 4)
        self._path_list.setShownCount(n_show)
        return super().resizeEvent(a0)


class QPathList(QtW.QGroupBox):
    pathClicked = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Recent")
        _layout = QtW.QVBoxLayout()
        _layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(_layout)
        paths = load_file_open_path()
        self._buttons: list[QClickableLabel] = []
        for path in reversed(paths):
            if path == "":
                continue
            btn = QClickableLabel(path)
            btn.setFixedHeight(_HEIGHT)
            btn.clicked.connect(lambda path=path: self.pathClicked.emit(path))
            _layout.addWidget(btn)
            self._buttons.append(btn)
        return None

    def setShownCount(self, n: int):
        """Set how many buttons to be shown."""
        _layout: QtW.QVBoxLayout = self.layout()
        n_wdt = _layout.count()
        if n > n_wdt:
            n = n_wdt

        for i in range(0, n):
            self._buttons[i].show()
        for i in range(n, n_wdt):
            self._buttons[i].hide()
        return None



from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING
from tabulous.widgets import TableViewer

if TYPE_CHECKING:
    from .widgets._mainwindow import TableViewerBase

CURRENT_VIEWER: TableViewerBase | None = None


def current_viewer() -> TableViewerBase:
    """Get the current table viewer widget."""
    global CURRENT_VIEWER
    if CURRENT_VIEWER is None:
        CURRENT_VIEWER = TableViewer()
    return CURRENT_VIEWER


def set_current_viewer(viewer: TableViewerBase) -> TableViewerBase:
    """Set a table viewer as the current one."""
    global CURRENT_VIEWER
    from .widgets._mainwindow import TableViewerBase

    if not isinstance(viewer, TableViewerBase):
        raise TypeError(f"Cannot set {type(viewer)} as the current viewer.")
    CURRENT_VIEWER = viewer
    return viewer


def read_csv(path: str | Path, *args, **kwargs) -> TableViewerBase:
    """Read CSV file and add it to the current viewer."""
    import pandas as pd

    df = pd.read_csv(path, *args, **kwargs)
    name = Path(path).stem
    viewer = current_viewer()
    viewer.add_table(df, name=name)
    return viewer


def read_excel(path: str | Path, *args, **kwargs) -> TableViewerBase:
    """Read Excel file and add all the sheets to the current viewer."""
    import pandas as pd

    df_dict: dict[str, pd.DataFrame] = pd.read_excel(
        path, *args, sheet_name=None, **kwargs
    )

    viewer = current_viewer()
    for sheet_name, df in df_dict.items():
        viewer.add_table(df, name=sheet_name)
    return viewer


def view_table(
    data,
    *,
    name: str | None = None,
    editable: bool = False,
    copy: bool = True,
) -> TableViewerBase:
    """View a table in the current viewer."""
    viewer = current_viewer()
    viewer.add_table(data, name=name, editable=editable, copy=copy)
    return viewer


def view_spreadsheet(
    data,
    *,
    name: str | None = None,
    editable: bool = True,
    copy: bool = True,
) -> TableViewerBase:
    """View a table as a spreadsheet in the current viewer."""
    viewer = current_viewer()
    viewer.add_spreadsheet(data, name=name, editable=editable, copy=copy)
    return viewer


def open_sample(
    sample_name: str,
    plugin_name: str = "seaborn",
) -> TableViewerBase:
    """Open a sample data."""
    viewer = current_viewer()
    viewer.open_sample(sample_name, plugin_name)
    return viewer



from __future__ import annotations
from functools import partial
from typing import (
    Any,
    Callable,
    Hashable,
    Iterator,
    MutableMapping,
    TYPE_CHECKING,
    TypeVar,
    Union,
)
import numpy as np
import pandas as pd
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt

if TYPE_CHECKING:
    from pandas.core.dtypes.dtypes import ExtensionDtype

    _DTypeLike = Union[np.dtype, ExtensionDtype]

_NAN_STRINGS = frozenset({"", "nan", "na", "n/a", "<na>", "NaN", "NA", "N/A", "<NA>"})


def _bool_converter(val: Any):
    if isinstance(val, str):
        if val in ("True", "1", "true"):
            return True
        elif val in ("False", "0", "false"):
            return False
        else:
            raise ValueError(f"Cannot convert {val} to bool.")
    else:
        return bool(val)


def _float_or_nan(x: Any):
    if x in _NAN_STRINGS:
        return float("nan")
    return float(x)


def _complex_or_nan(x: Any):
    if x in _NAN_STRINGS:
        return float("nan")
    return complex(x)


_INTERVAL_PARENTHESIS = {
    ("[", "]"): "both",
    ("[", ")"): "left",
    ("(", "]"): "right",
    ("(", ")"): "neither",
}


def _to_interval(x: Any) -> pd.Interval:
    """Convert an object into a pd.Interval object."""
    if isinstance(x, str):
        if x == "":
            return pd.NA
        elif len(x) < 5:
            raise ValueError(f"Cannot conver string {x!r} to an interval.")
        left_par = x[0]
        right_par = x[-1]
        rest = x[1:-1]
        values = eval(f"({rest})", {"pd": pd}, {})
        closed = _INTERVAL_PARENTHESIS.get((left_par, right_par))
        return pd.Interval(*values, closed=closed)
    elif isinstance(x, pd.Interval):
        return x
    elif hasattr(x, "__iter__"):
        return pd.Interval(*x)
    elif isinstance(x, slice):
        if x.start is None or x.stop is None or x.step not in (1, None):
            raise ValueError(f"Slice {x} not interpretable.")
        return pd.Interval(x.start, x.stop)
    else:
        raise TypeError(f"Cannot convert {type(x)} to pd.Interval.")


_DTYPE_CONVERTER = {
    "i": int,
    "f": _float_or_nan,
    "u": int,
    "b": _bool_converter,
    "U": str,
    "c": _complex_or_nan,
    "M": pd.to_datetime,
    "m": pd.to_timedelta,
}

_GET_CONVERTER_CACHE: dict[Hashable, Callable[[Any], Any]] = {}


def _IDENTITY_CONVERTER(e):
    return e


def get_converter(dtype: _DTypeLike | str) -> Callable[[Any], Any]:
    """Get a scalar value converter function for the given dtype"""
    # try to use the dtype as a key to the cache
    _is_hashable = hasattr(dtype, "__hash__")
    if _is_hashable and dtype in _GET_CONVERTER_CACHE:
        return _GET_CONVERTER_CACHE[dtype]

    _dtype = get_dtype(dtype)
    _kind = _dtype.kind
    if _kind != "O":
        out = _DTYPE_CONVERTER[_kind]
    else:
        if isinstance(_dtype, pd.IntervalDtype):
            out = _to_interval
        elif isinstance(_dtype, pd.PeriodDtype):
            out = partial(pd.Period, freq=_dtype.freq)
        elif isinstance(_dtype, pd.CategoricalDtype):
            out = get_converter(_dtype.categories.dtype)
        else:
            out = _IDENTITY_CONVERTER

    if _is_hashable:
        _GET_CONVERTER_CACHE[dtype] = out
    return out


def get_converter_from_type(tp: type | str) -> Callable[[Any], Any]:
    import datetime

    if isinstance(tp, str):
        ns = {"pd": pd, "datetime": datetime, "timedelta": datetime.timedelta}
        try:
            tp = eval(tp, ns, {})
        except NameError:
            return _IDENTITY_CONVERTER

    if tp is int:
        kind = "i"
    elif tp is float:
        kind = "f"
    elif tp is str:
        kind = "U"
    elif tp is bool:
        kind = "b"
    elif tp is complex:
        kind = "c"
    elif tp in (datetime, datetime.datetime, pd.Timestamp):
        kind = "M"
    elif tp in (datetime.timedelta, pd.Timedelta):
        kind = "m"
    elif tp is pd.Period:
        return pd.Period
    elif tp is pd.Interval:
        return _to_interval
    else:
        return _IDENTITY_CONVERTER
    return _DTYPE_CONVERTER[kind]


class DefaultValidator:
    """
    The default validator function.

    This class is a simple wrapper around a callable. Spreadsheet needs to know if
    a validator is the default one when dtype is changed.
    """

    def __init__(self, dtype: Any):
        self._dtype = get_dtype(dtype)
        self._converter = _DTYPE_CONVERTER.get(self._dtype.kind, lambda x: None)

    def __call__(self, value: Any) -> None:
        self._converter(value)
        return None

    def __repr__(self) -> str:
        return f"DefaultValidator[{self._dtype.name}]"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DefaultValidator):
            return False
        return self._dtype == other._dtype


def get_dtype(dtype: Any):
    """Get pandas dtype."""
    from pandas.core.dtypes.common import pandas_dtype

    return pandas_dtype(dtype)


class QDtypeWidget(QtW.QTreeWidget):
    """A tree view widget of supported dtypes."""

    _DTYPES = [
        "unset",
        "int",
        "uint",
        "float",
        "complex",
        "bool",
        "time",
        "others",
    ]

    _SUB_DTYPES = {
        "int": ["int8", "int16", "int32", "int64"],
        "uint": ["uint8", "uint16", "uint32", "uint64"],
        "float": ["float16", "float32", "float64"],
        "complex": ["complex64", "complex128"],
        "time": ["datetime64", "timedelta64", "period"],
        "others": ["category", "interval", "string", "object"],
    }

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setHeaderHidden(True)

        for dtype in self._DTYPES:
            item = QtW.QTreeWidgetItem([dtype])
            self.addTopLevelItem(item)
            if sub := self._SUB_DTYPES.get(dtype, None):
                item.addChildren(QtW.QTreeWidgetItem([s]) for s in sub)

        self.setSelectionMode(QtW.QTreeWidget.SelectionMode.SingleSelection)

    def dtypeText(self) -> str:
        """Get the selected dtype as a text."""
        return self.currentItem().text(0)

    @classmethod
    def requestValue(cls, parent=None) -> tuple[str, bool] | None:
        """Ask the user to select a dtype."""
        self = cls()
        dlg = QtW.QDialog(parent)
        dlg.setLayout(QtW.QVBoxLayout())
        dlg.layout().addWidget(QtW.QLabel("dtype"))
        dlg.layout().addWidget(self)

        validation_checkbox = QtW.QCheckBox("Data validation")
        validation_checkbox.setChecked(True)
        validation_checkbox.setToolTip(
            "Check to set the data validator for the data type."
        )
        dlg.layout().addWidget(validation_checkbox)

        fmt_checkbox = QtW.QCheckBox("Data formatting")
        fmt_checkbox.setChecked(True)
        fmt_checkbox.setToolTip(
            "Check to set the default data formatter for the data type."
        )
        dlg.layout().addWidget(fmt_checkbox)

        _Btn = QtW.QDialogButtonBox.StandardButton
        _btn_box = QtW.QDialogButtonBox(
            _Btn.Ok | _Btn.Cancel,
            Qt.Orientation.Horizontal,
        )
        _btn_box.accepted.connect(dlg.accept)
        _btn_box.rejected.connect(dlg.reject)
        dlg.layout().addWidget(_btn_box)

        @self.doubleClicked.connect
        def _on_double_click(index: QtCore.QModelIndex):
            item = self.itemFromIndex(index)
            if item.childCount() == 0:
                dlg.accept()
            return

        dlg.setWindowTitle("Select a dtype")
        if dlg.exec():
            out = (
                self.dtypeText(),
                validation_checkbox.isChecked(),
                fmt_checkbox.isChecked(),
            )
        else:
            out = None
        return out


_K = TypeVar("_K", bound=Hashable)
_V = TypeVar("_V", bound="_DTypeLike")


class DTypeMap(MutableMapping[_K, _V]):
    """
    Mapping storage of dtypes.

    The dtype map cannot be simply represented as a dict because datetime, timedelta
    and complex must be passed differently in pd.read_csv.
    """

    def __init__(self) -> None:
        self._dict: dict[Hashable, _DTypeLike] = {}
        self._datetime_dict: dict[Hashable, _DTypeLike] = {}
        self._need_parsing_dict: dict[Hashable, _DTypeLike] = {}

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"{clsname}{dict(**self)!r}"

    def __getitem__(self, key: _K) -> _V:
        out = self._dict.get(key, None)
        if out is None:
            out = self._datetime_dict.get(key, None)
            if out is None:
                out = self._need_parsing_dict[key]
        return out

    def __setitem__(self, key: _K, dtype: _V) -> None:
        if dtype.kind not in ("M", "m", "c"):
            if dtype == "interval" or dtype == "period":
                self._need_parsing_dict[key] = dtype
            else:
                self._dict[key] = dtype
        elif dtype.kind == "M":
            self._datetime_dict[key] = dtype
        else:
            self._need_parsing_dict[key] = dtype

    def __delitem__(self, key: _K) -> None:
        self._dict.pop(key, None) or self._datetime_dict.pop(
            key, None
        ) or self._need_parsing_dict.pop(key, None)
        return None

    def __iter__(self) -> Iterator[_K]:
        yield from iter(self._dict)
        yield from iter(self._datetime_dict)
        yield from iter(self._need_parsing_dict)

    def __len__(self) -> int:
        return len(self._dict) + len(self._datetime_dict) + len(self._need_parsing_dict)

    def as_pandas_kwargs(self) -> dict[str, Any]:
        """Create a dict ready for being passed to ``pd.read_csv``."""
        kwargs = {"dtype": self._dict}
        if self._datetime_dict:
            kwargs.update(
                parse_dates=list(self._datetime_dict.keys()),
                infer_datetime_format=True,
            )
        if self._need_parsing_dict:
            kwargs.update(
                converters={
                    k: get_converter(v) for k, v in self._need_parsing_dict.items()
                },
            )
        return kwargs

    def try_convert(self, key: _K, value: Any) -> Any:
        """Convert value according to the dtype, if registered."""
        if dtype := self.get(key, None):
            return get_converter(dtype)(value)
        return value

    def copy(self) -> DTypeMap:
        new = DTypeMap()
        new._dict = self._dict.copy()
        new._datetime_dict = self._datetime_dict.copy()
        new._need_parsing_dict = self._need_parsing_dict.copy()
        return new


_NANS = frozenset({np.nan, pd.NA, float("nan")})


def isna(val: Any):
    # NOTE: pd.isna is slow.
    return val in _NANS



from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any
from datetime import datetime, timedelta
from psygnal import Signal
from magicgui.widgets import (
    Container,
    PushButton,
    DateTimeEdit,
    LineEdit,
    ComboBox,
)
from tabulous._magicgui import SelectionWidget, TimeDeltaEdit

import pandas as pd

if TYPE_CHECKING:
    from magicgui.widgets.bases import ValueWidget


class _CheckedWidget(Container):
    check_changed = Signal(bool)

    def __init__(self, widget: ValueWidget, checked: bool = True, **kwargs):
        from tabulous._magicgui import ToggleSwitch

        self._cbox = ToggleSwitch(value=checked)
        self._child_widget = widget
        super().__init__(
            widgets=[self._cbox, widget],
            layout="horizontal",
            labels=False,
            name=widget.name,
            label=widget.label,
            **kwargs,
        )
        self.margins = (0, 0, 0, 0)
        self._cbox.changed.connect(self.check_changed.emit)

    @property
    def value(self) -> Any:
        return self._child_widget.value

    @value.setter
    def value(self, val):
        self._child_widget.value = val

    @property
    def checked(self):
        return self._cbox.value

    @checked.setter
    def checked(self, val):
        self._cbox.value = val


class _RangeDialog(Container):
    def __init__(self):
        selection = SelectionWidget(
            format="iloc", allow_out_of_bounds=True, name="Selection"
        )
        start, end = self._prep_start_and_end()
        start = _CheckedWidget(start)
        end = _CheckedWidget(end)
        freq = _CheckedWidget(self._prep_freq(), checked=False, enabled=False)

        start.check_changed.connect(lambda: self._on_check_state_changed(start))
        end.check_changed.connect(lambda: self._on_check_state_changed(end))
        freq.check_changed.connect(lambda: self._on_check_state_changed(freq))
        call_btn = PushButton(text="Run")
        super().__init__(widgets=[selection, start, end, freq, call_btn])
        self._selection = selection
        self._start = start
        self._end = end
        self._freq = freq
        self._call_btn = call_btn

        self.called = self._call_btn.changed
        self.called.connect(self.close)

    def _on_check_state_changed(self, widget: _CheckedWidget):
        for _w in (self._start, self._end, self._freq):
            if _w is widget:
                assert not _w.checked
            else:
                with _w.check_changed.blocked():
                    _w.checked = True
            _w.enabled = _w.checked

    def _prep_start_and_end(self) -> tuple[_CheckedWidget, _CheckedWidget]:
        raise NotImplementedError()

    def _prep_freq(self) -> _CheckedWidget:
        raise NotImplementedError()

    def get_range(self, start=None, end=None, periods=None, freq=None):
        raise NotImplementedError()

    def _get_params(self):
        start = self._start.value if self._start.checked else None
        end = self._end.value if self._end.checked else None
        freq = self._freq.value if self._freq.checked else None
        return start, end, freq

    def get_value(self, df: pd.DataFrame) -> tuple[slice, slice, pd.Index]:
        rsl, csl = self._selection.value.as_iloc_slices(df)
        if csl.start != csl.stop - 1:
            raise ValueError("Selection must be a single column")
        periods = rsl.stop - rsl.start
        start, end, freq = self._get_params()
        data = self.get_range(start=start, end=end, periods=periods, freq=freq)
        return rsl, csl, data


class _TimeRangeDialog(_RangeDialog):
    def _prep_freq(self) -> _CheckedWidget:
        return TimeDeltaEdit(name="freq", value=timedelta(days=1))


class DateRangeDialog(_TimeRangeDialog):
    def _prep_start_and_end(self):
        start = DateTimeEdit(name="start", value=datetime(2000, 1, 1))
        end = DateTimeEdit(name="stop", value=datetime(2000, 12, 31))
        return start, end

    def get_range(self, start=None, end=None, periods=None, freq=None):
        return pd.date_range(start=start, end=end, periods=periods, freq=freq)


class TimeDeltaRangeDialog(_TimeRangeDialog):
    def _prep_start_and_end(self):
        start = TimeDeltaEdit(name="start", value=timedelta(seconds=0))
        end = TimeDeltaEdit(name="stop", value=timedelta(seconds=60))
        return start, end

    def get_range(self, start=None, end=None, periods=None, freq=None):
        return pd.timedelta_range(start=start, end=end, periods=periods, freq=freq)


class PeriodRangeDialog(_TimeRangeDialog):
    def _prep_start_and_end(self):
        start = DateTimeEdit(name="start", value=datetime(2000, 1, 1))
        end = DateTimeEdit(name="stop", value=datetime(2000, 12, 31))
        return start, end

    def _prep_freq(self) -> _CheckedWidget:
        return LineEdit(name="freq", value="1d", tooltip="Period frequency")

    def get_range(self, start=None, end=None, periods=None, freq=None):
        return pd.period_range(start=start, end=end, periods=periods, freq=freq)


class IntervalRangeDialog(_RangeDialog):
    def __init__(self):
        super().__init__()
        self._closed = ComboBox(
            choices=[
                ("[a, b)", "left"),
                ("(a, b]", "right"),
                ("[a, b]", "both"),
                ("(a, b)", "neither"),
            ],
            name="close_state",
        )

    def _prep_start_and_end(self):
        start = LineEdit(name="start", value="0.0")
        end = LineEdit(name="stop", value="1.0")
        return start, end

    def _prep_freq(self):
        return LineEdit(name="freq", value="0.1")

    def get_range(self, start=None, end=None, periods=None, freq=None):
        return pd.interval_range(
            start=ast.literal_eval(start) if start is not None else None,
            end=ast.literal_eval(end) if end is not None else None,
            periods=periods,
            freq=ast.literal_eval(freq) if freq is not None else None,
            closed=self._closed.value,
        )



from __future__ import annotations
from typing import TYPE_CHECKING
from . import _dialogs
from tabulous._magicgui import ToggleSwitchSelect

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase


def activate_left(viewer: TableViewerBase):
    """Activate left table."""
    num = len(viewer.tables)
    if num == 0:
        return None
    src = viewer.current_index
    if src == 0:
        return None
    viewer.current_index = src - 1
    return viewer.native.setCellFocus()


def activate_right(viewer: TableViewerBase):
    """Activate right table."""
    num = len(viewer.tables)
    if num == 0:
        return None
    src = viewer.current_index
    if src == num - 1:
        return None
    viewer.current_index = src + 1
    return viewer.native.setCellFocus()


def swap_tab_with_left(viewer: TableViewerBase):
    """Swap the current tab with the left one"""
    num = len(viewer.tables)
    if num == 0:
        return None
    src = viewer.current_index
    if src == 0:
        return None
    viewer.tables.move(src, src - 1)


def swap_tab_with_right(viewer: TableViewerBase):
    """Swap the current tab with the right one"""
    num = len(viewer.tables)
    if num == 0:
        return None
    src = viewer.current_index
    if src == num - 1:
        return None
    viewer.tables.move(src + 1, src)


def tile_tables(viewer: TableViewerBase):
    """Tile tabs"""
    choices = [(table.name, idx) for idx, table in enumerate(viewer.tables)]
    out = _dialogs.choose_multiple(
        choices={"choices": choices, "widget_type": ToggleSwitchSelect},
        parent=viewer.native,
    )
    if out:
        viewer.tables.tile(out)


def tile_with_adjacent_table(viewer: TableViewerBase):
    """Tile adjacent tabs"""
    num = len(viewer.tables)
    if num < 2:
        return None
    idx = viewer.current_index
    if idx < num - 1:
        indices = [idx, idx + 1]
    else:
        indices = [idx - 1, idx]
    all_indices = []
    for i in indices:
        all_indices.extend(viewer.native._tablestack.tiledIndices(i))

    all_indices = list(set(all_indices))
    return viewer.tables.tile(all_indices, orientation="horizontal")


def untile_table(viewer: TableViewerBase):
    """Untile current tab"""
    viewer.tables.untile(viewer.current_index)


def rename_tab(viewer: TableViewerBase):
    """Rename tab."""
    idx = viewer.current_index
    if idx is None:
        raise ValueError("No tab found.")
    viewer.native._tablestack.enterEditingMode(idx)


def delete_tab(viewer: TableViewerBase) -> None:
    """Delete current tab."""
    idx = viewer.current_index
    viewer.tables.del_checked(idx)



from __future__ import annotations
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt

if TYPE_CHECKING:
    from ._tabwidget import QTabbedTableStack

# https://stackoverflow.com/questions/50578661/how-to-implement-vertical-tabs-in-qt


class QTabulousTabBar(QtW.QTabBar):
    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setMouseTracking(True)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

    if TYPE_CHECKING:

        def parentWidget(self) -> QTabbedTableStack:
            ...


class _QSideBar(QTabulousTabBar):
    _ROTATION_DEGREE: int

    def tabSizeHint(self, index: int) -> QtCore.QSize:
        s = super().tabSizeHint(index)
        s.transpose()
        s.setWidth(120)
        return s

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtW.QStylePainter(self)
        opt = QtW.QStyleOptionTab()

        TabBarTabShape = QtW.QStyle.ControlElement.CE_TabBarTabShape
        TabBarTabLabel = QtW.QStyle.ControlElement.CE_TabBarTabLabel
        for i in range(self.count()):
            self.initStyleOption(opt, i)
            painter.drawControl(TabBarTabShape, opt)

            painter.save()

            s = opt.rect.size()
            s.transpose()
            r = QtCore.QRect(QtCore.QPoint(), s)
            r.moveCenter(opt.rect.center())
            opt.rect = r

            tab_rect = self.tabRect(i)
            c = tab_rect.center()
            painter.translate(c)
            painter.rotate(self._ROTATION_DEGREE)
            painter.translate(-c)
            painter.drawControl(TabBarTabLabel, opt)
            painter.restore()

        QtW.QWidget.paintEvent(self, event)
        return None


class QLeftSideBar(_QSideBar):
    _ROTATION_DEGREE = 90


class QRightSideBar(_QSideBar):
    _ROTATION_DEGREE = 270



from __future__ import annotations
from typing import TYPE_CHECKING
from . import _dialogs, _utils
from tabulous._magicgui import ToggleSwitchSelect

if TYPE_CHECKING:
    from tabulous.widgets._mainwindow import TableViewerBase

SUMMARY_CHOICES = [
    "mean",
    "median",
    "std",
    "sem",
    "min",
    "max",
    "sum",
    "nunique",
    "dtype",
]


def summarize_table(viewer: TableViewerBase):
    """Summarize table data"""
    table = _utils.get_table(viewer)
    out = _dialogs.summarize_table(
        table={"bind": table},
        methods={"choices": SUMMARY_CHOICES, "widget_type": ToggleSwitchSelect},
        parent=viewer._qwidget,
    )
    if out is not None:
        df, new = out
        if new:
            viewer.add_table(df, name=f"{table.name}-summary")
        else:
            from tabulous._qt._table import QTableLayer

            qtable = QTableLayer(data=df)
            qtable.setZoom(0.8)
            table.add_side_widget(qtable, name="summary")


def show_filter_widget(viewer: TableViewerBase):
    """Show filter widget."""
    return viewer._qwidget._tablestack.openFilterDialog()


def show_eval_widget(viewer: TableViewerBase):
    """Evaluate a Python expression"""
    return viewer._qwidget._tablestack.openEvalDialog()


def show_optimizer_widget(viewer: TableViewerBase):
    """Open optimizer widget"""
    from ._optimizer import OptimizerWidget

    tablestack = viewer._qwidget._tablestack
    ol = tablestack._overlay
    ol.show()

    ol.addWidget(OptimizerWidget.new().native)
    ol.setTitle("Optimization")


def show_stats_widget(viewer: TableViewerBase):
    """Open scipy.stats widget"""
    from ._statistics import StatsTestDialog

    dlg = StatsTestDialog()
    dlg.native.setParent(viewer._qwidget, dlg.native.windowFlags())
    dlg.show()


def show_sklearn_widget(viewer: TableViewerBase):
    """Open scikit-learn widget"""
    from ._sklearn import SkLearnContainer

    tablestack = viewer._qwidget._tablestack
    ol = tablestack._overlay
    ol.show()

    ol.addWidget(SkLearnContainer.new().native)
    ol.setTitle("scikit-learn analysis")
